import json
import os
from tqdm import tqdm
import numpy as np
import pandas as pd

def save_results(data, output_path):
    js_file = f"{output_path}/FID_output.json"
    os.makedirs(os.path.dirname(js_file), exist_ok=True)
    try:
        with open(js_file, "w") as f:
            json.dump(clean_for_json(data), f, indent=4)
        # tqdm.write(f"Output structure saved to:\n{js_file}")
    except Exception as e:
        tqdm.write("Error saving JSON:", e)

# def load_results(output_path, filename="FID_output.json"):
#     js_file = os.path.join(output_path, filename)
#     if os.path.exists(js_file):
#         try:
#             with open(js_file, "r") as f:
#                 return json.load(f)
#         except Exception as e:
#             tqdm.write("Error loading existing JSON:", e)
#     return None

def load_results(results_path, list_samples=False, list_processed=False):
    """
    Try to load FID_output.json from output_path.
    If it doesn’t exist, return None.
    Otherwise return the dict, rebuilding any Raw Data dicts into DataFrames.
    """
    js_file = os.path.join(results_path)
    if not os.path.exists(js_file):
        return None

    with open(js_file, "r") as f:
        data = json.load(f)

    # rebuild Raw Data dicts into DataFrames
    for sample in data.get("Samples", {}).values():
        raw = sample.get("Raw Data")
        if isinstance(raw, dict):
            sample["Raw Data"] = pd.DataFrame(raw)
    if list_samples:
        for key in data['Samples'].keys():
            print(key)
    if list_processed:
        key = []
        for x in data['Samples'].keys():
            if 'Processed Data' in data['Samples'][x].keys():
                key.append(x)
        print(key)
    return data


def clean_for_json(obj):
    if isinstance(obj, (np.integer, np.floating)):
        return obj.item()
    elif isinstance(obj, (np.ndarray, pd.Series, list, tuple)):
        return [clean_for_json(el) for el in obj]
    elif isinstance(obj, dict):
        return {str(k): clean_for_json(v) for k, v in obj.items()}
    else:
        try:
            json.dumps(obj)  # test if serializable
            return obj
        except (TypeError, OverflowError):
            return str(obj)  # fallback

def delete_samples(json_path: str, to_delete: list[str]) -> dict:
    """
    Load an FID integration JSON, delete any samples whose key contains
    one of the given substrings, and overwrite the file.
    """
    if not os.path.isfile(json_path):
        raise FileNotFoundError(f"No such file: {json_path!r}")
    with open(json_path, "r") as f:
        data = json.load(f)

    # grab the actual dict
    samples: dict = data.get("Samples", {})

    # make a static list of the keys so we can pop safely
    all_keys = list(samples.keys())

    for fragment in to_delete:
        # print(f"Looking for samples containing: {fragment!r}")
        # find matches in that static list
        matches = [k for k in all_keys if fragment in k]
        if not matches:
            tqdm.write(f"Could not delete any samples matching '{fragment}'")
            continue

        for key in matches:
            # actually remove from the dict
            samples.pop(key, None)
            print(f"Deleted sample: {key}  (matched '{fragment}')")

    # write back
    with open(json_path, "w") as f:
        json.dump(data, f, indent=4)


"""
/src/chromatopy/FID/__init__.py
"""

from .FID_integration import integration
from .peak_labels_editor import peak_label_editor
from .FID_General import plot_chromatogram, plot_chromatogram_cluster
from .bouqueter import cluster
from .Tools import save_results, load_results, delete_samples

__all__ = [
	 "integration",
	 "peak_label_editor",
	 "plot_chromatogram",
    "plot_chromatogram_cluster",
	 "cluster",
	 "save_results",
	 "load_results",
	 "delete_samples"]

name = "FID"



import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import find_peaks
import hdbscan
from tqdm import tqdm

from .FID_Integration_functions import *
from .import_data import import_data

import os
import shutil


def extract_peak_times_from_sample(sample_df, time_range=(2, 20), prominence=1):
    time = sample_df['Time (min)'].values
    signal = sample_df['Value (pA)'].values
    mask = (time >= time_range[0]) & (time <= time_range[1])
    if not np.any(mask):
        return np.array([])

    time_in_range = time[mask]
    signal_in_range = signal[mask]
    norm_signal = signal_in_range / np.max(signal_in_range)
    _, min_peak_amp = baseline(time_in_range, norm_signal)
    peaks, _ = find_peaks(norm_signal, prominence=prominence, height = min_peak_amp)
    return time_in_range[peaks]

def peak_time_distance_with_threshold(a, b, threshold):
    if len(a) == 0 or len(b) == 0:
        return 1.0  # maximum dissimilarity

    def count_unmatched_and_error(x, y):
        matched = []
        unmatched = 0
        total_error = 0
        for xi in x:
            diffs = np.abs(y - xi)
            if np.any(diffs <= threshold):
                closest = np.min(diffs)
                total_error += closest
                matched.append(xi)
            else:
                unmatched += 1
        return unmatched, total_error

    unmatched_a, error_a = count_unmatched_and_error(a, b)
    unmatched_b, error_b = count_unmatched_and_error(b, a)

    total_peaks = len(a) + len(b)
    total_unmatched = unmatched_a + unmatched_b
    total_error = error_a + error_b

    # Normalize to get a distance between 0 and 1
    normalized_distance = (total_unmatched + total_error / threshold) / total_peaks
    return normalized_distance

def compute_timing_distance_matrix(peak_time_list, threshold=0.2):
    n = len(peak_time_list)
    dist = np.zeros((n, n))
    for i in range(n):
        for j in range(i + 1, n):
            d = peak_time_distance_with_threshold(peak_time_list[i], peak_time_list[j], threshold=threshold)
            dist[i, j] = dist[j, i] = d
    return dist


def cluster_samples_by_peak_timing(data_dict, min_cluster_size, min_samples, time_range=(2, 20), prominence=10):
    sample_names = list(data_dict['Samples'].keys())
    peak_times_list = []

    for name in tqdm(sample_names, desc="Processing chromatography similairty"):
        df = data_dict['Samples'][name]['raw data']

        time_col = next((col for col in df.columns if 'Time (min)' in col), None)
        signal_col = next((col for col in df.columns if 'Value (pA)' in col), None)

        if 'Time (min)' in df.columns and 'Value (pA)' in df.columns:
            peak_times = extract_peak_times_from_sample(df[['Time (min)', 'Value (pA)']], time_range, prominence)
            peak_times_list.append(peak_times)
        else:
            tqdm.write(f"Skipping {name}: missing 'Time (min)' or 'Value (pA)' columns: {df.columns}")
            peak_times_list.append(np.array([]))

    dist_matrix = compute_timing_distance_matrix(peak_times_list)
    clusterer = hdbscan.HDBSCAN(metric='precomputed',min_cluster_size=min_cluster_size, min_samples=min_samples)
    labels = clusterer.fit_predict(dist_matrix)
    cluster_labels = np.unique(labels)

    # Add cluster label to each sample's dictionary
    for name, label in zip(sample_names, labels):
        data_dict['Samples'][name]['cluster'] = int(label)
    return cluster_labels

def move_files_to_cluster_subfolder(file_names, cluster_var, source_folder):
    # Create destination subfolder
    dest_folder = os.path.join(source_folder, f"cluster_{cluster_var}")
    os.makedirs(dest_folder, exist_ok=True)

    # Normalize file names (remove extension for matching)
    for filename in os.listdir(source_folder):
        full_path = os.path.join(source_folder, filename)
        if os.path.isfile(full_path):
            name_without_ext = os.path.splitext(filename)[0]
            if name_without_ext in file_names:
                shutil.move(full_path, os.path.join(dest_folder, filename))
                tqdm.write(f"Moved {filename} to cluster_{cluster_var}/")

    tqdm.write(f"All matching files moved to {dest_folder}.")
    
def cluster_figures(data):
    cluster_labels = set()
    for key in data['Samples'].keys():
        cluster_labels.add(data['Samples'][key]['cluster'])
    for cl in cluster_labels:
        fig=plt.figure()
        for key in data['Samples'].keys():
            if data['Samples'][key]['cluster'] != cl: continue
            plt.plot(data['Samples'][key]['raw data']['Time (min)'], data['Samples'][key]['raw data']['Value (pA)'], c= 'k')
        plt.xlabel("Time (min)")
        plt.ylabel(f"Value (pA)\n{cl}")
        plt.show()
        
def get_cluster_labels(data):
    cluster_labels = set()
    for key in data['Samples'].keys():
        cluster_labels.add(data['Samples'][key]['cluster'])
    return cluster_labels

def plot_chromatogram(time_window=None):
    """
    Plot and save chromatograms of all FID samples.

    Parameters
    ----------
    time_window : list of two floats, optional
        [xmin, xmax] time limits (in same units as input) to restrict the plot.
        If None, the full trace is plotted.
    """
    # 1) load data & get the raw folder path
    result = import_data()
    data = result['data_dict']
    time_column = result["time_column"]
    signal_column = result["signal_column"]
    folder_path = result["folder_path"]
    output_path = result["output_path"]
    figures_path = result["figures_path"]
    
    # 2) make (or reuse) the output folder
    chromatograms_dir = os.path.join(folder_path, "chromatograms")
    os.makedirs(chromatograms_dir, exist_ok=True)

    # 3) loop and save each sample’s trace
    for key, sample in data['Samples'].items():
        df = pd.DataFrame({
            'x': sample['Raw Data'][time_column],
            'y': sample['Raw Data'][signal_column]
        })
        if time_window is not None:
            xmin, xmax = time_window
            df = df.loc[(df.x >= xmin) & (df.x <= xmax)]

        plt.figure()
        plt.plot(df.x, df.y, color='k')
        plt.xlabel(time_column)
        plt.ylabel(signal_column)

        out_file = os.path.join(chromatograms_dir, f"{key}.png")
        plt.savefig(out_file, dpi=300)
        plt.close()
    return data, chromatograms_dir, output_path, folder_path
        
def plot_chromatogram_cluster(data, time_window=[4,30]):
    """
    Function for plotting the categorized chromatograms.

    Parameters
    ----------
    data : TYPE
        DESCRIPTION.
    time_window : TYPE, optional
        DESCRIPTION. The default is [4,30].

    Returns
    -------
    None.

    """
    cluster_n = 0
    for key in data['Samples'].keys():
        if 'cluster' not in data['Samples'][key].keys():
            tqdm.write("Data does not appear to contain cluster labels. Please run clusterer().")
            return
        else:
            if data['Samples'][key]['cluster'] > cluster_n:
                cluster_n = data['Samples'][key]['cluster']
    for i in range(1,cluster_n+1):
        fig = plt.figure()
        count = 0
        for key in data['Samples'].keys():
            # if group_assignments[key] ==val:
            if data['Samples'][key]['cluster']==i:
                count +=1
                temp = data['Samples'][key]['Raw Data']
                temp = temp[temp['Time (min)']>4]
                plt.plot(temp['Time (min)'], temp[f'Value (pA)'], c='k', alpha = 0.4)
        plt.text(x=0.05, y = 0.95, s=f"n = {count}", transform=plt.gca().transAxes, 
                 fontsize=12, verticalalignment='top')
        plt.ylabel(f"Signal (pA)\nCluster {i}")
        plt.xlabel("Time (min)")
        plt.show()


# This file defines: peak_label_editor()
import sys
import json
import os
from PyQt5.QtWidgets import (
    QApplication, QDialog, QVBoxLayout, QLabel, QLineEdit,
    QPushButton, QHBoxLayout, QMessageBox
)

DEFAULT_LABELS = {
    "Peak Labels": ["C16", "C17", "C18", "C19", "C20Z2",
                    "C20", "C21", "C22", "C23", "C24",
                    "C25", "C26", "C27", "C28", "C29",
                    "C30", "C31", "C32"],
    "x limits": [3, 17]
}

MODULE_DIR = os.path.dirname(__file__)
DEFAULT_JSON = os.path.join(MODULE_DIR, "peak_labels.json")

def load_peak_labels(json_path: str = DEFAULT_JSON):
    if os.path.exists(json_path):
        with open(json_path, 'r') as f:
            return json.load(f)
    return DEFAULT_LABELS

def save_peak_labels(data, json_path: str = DEFAULT_JSON):
    # debug print so you can see where it goes:
    # print(f"⤷ writing peak_labels.json to: {json_path!r}")
    with open(json_path, 'w') as f:
        json.dump(data, f, indent=4)

class PeakLabelEditor(QDialog):
    def __init__(self, json_path: str = DEFAULT_JSON):
        super().__init__()
        self.json_path = json_path
        self.label_dict = load_peak_labels(json_path)
        self.init_ui()

    def init_ui(self):
        layout = QVBoxLayout()
        layout.addWidget(QLabel("Peak Labels (comma-separated):"))
        self.labels_input = QLineEdit(", ".join(self.label_dict.get("Peak Labels", [])))
        layout.addWidget(self.labels_input)

        layout.addWidget(QLabel("X limits (e.g., 5, 17):"))
        self.xlims_input = QLineEdit(", ".join(str(x) for x in self.label_dict.get("x limits", [5, 17])))
        layout.addWidget(self.xlims_input)

        save_btn = QPushButton("Save")
        save_btn.clicked.connect(self.save_and_exit)
        btn_layout = QHBoxLayout()
        btn_layout.addStretch()
        btn_layout.addWidget(save_btn)
        layout.addLayout(btn_layout)

        self.setLayout(layout)

    def save_and_exit(self):
        try:
            labels = [l.strip() for l in self.labels_input.text().split(",") if l.strip()]
            xlims = [float(x.strip()) for x in self.xlims_input.text().split(",")]
            if len(xlims) != 2:
                raise ValueError("x limits must have exactly two numbers.")
            new_data = {"Peak Labels": labels, "x limits": xlims}
            save_peak_labels(new_data, self.json_path)
            QMessageBox.information(self, "Saved", f"Saved to {self.json_path}")
            self.accept()
        except Exception as e:
            QMessageBox.critical(self, "Error", f"Failed to save: {str(e)}")

def peak_label_editor(json_path: str = DEFAULT_JSON):
    app = QApplication.instance() or QApplication(sys.argv)
    editor = PeakLabelEditor(json_path)
    editor.exec_()



#!/chromatoPy/src/chromatopy/FID/import_data.py
import os
from tqdm import tqdm
import re
import pandas as pd
import json

def import_data(): # folder_path=None):
    folder_path = input("Provide folder containing .txt files: ")
    folder_path = folder_path.strip('\'"')

    # List all .txt files
    txt_files = [f for f in os.listdir(folder_path) if f.lower().endswith(".txt")]
    if not txt_files:
        tqdm.write(f"No .txt files found in {folder_path}. Aborting.")
        raise SystemExit

    output_path, figures_path = create_output_folders(folder_path)

    no_time_col = []
    no_signal_col = []
    data_dict = {}
    data_dict["Samples"] = {}

    for filename in txt_files:
        file_path = os.path.join(folder_path, filename)

        with open(file_path, 'r', encoding='utf-8', errors='ignore') as f:
            lines = f.readlines()

        # Find the line where the table begins
        table_start = None
        for i, line in enumerate(lines):
            if re.search(r'(?i)^\s*Chromatogram Data\s*:', line):
                table_start = i + 1  # The actual table header is the next line
                break

        if table_start is None or table_start + 1 >= len(lines):
            tqdm.write(f"Could not find table start in {filename}")
            print(f"Could not find table start in {filename}")
            continue

        # Read headers
        headers = lines[table_start].strip().split('\t')
        data_lines = lines[table_start + 1:]

        # Read into DataFrame
        try:
            df = pd.DataFrame([l.strip().split('\t') for l in data_lines if l.strip() != ''], columns=headers)
        except Exception as e:
            tqdm.write(f"Failed to parse table in {filename}: {e}")
            print(f"Failed to parse table in {filename}: {e}")
            continue

        # Header matching
        time_keywords = ['time', 'min', 'sec', 'second', 'minute']
        signal_keywords = ['signal', 'value', 'intensity', 'amplitude', '(pa)', '(a)']
        headers = lines[table_start].strip().split('\t')
        header_map = {h.lower(): h for h in headers}

        time_column = next((header_map[h] for h in header_map if any(key in h for key in time_keywords)), None)
        has_time = time_column is not None
        signal_column = next((header_map[h] for h in header_map if any(key in h for key in signal_keywords)), None)
        has_signal = signal_column is not None

        # Numeric dataframe
        df[time_column] = pd.to_numeric(df[time_column], errors='coerce')
        df[signal_column] = pd.to_numeric(df[signal_column], errors='coerce')
        df[signal_column] = df[signal_column]#.fillna(0)

        if not has_time:
            no_time_col.append(filename)
        if not has_signal:
            no_signal_col.append(filename)
        metadata = ''.join(lines[:table_start - 1])
        parsed_metadata = parse_metadata_block(metadata)

        # Store in dictionary
        data_dict['Samples'][filename.replace(".txt", "")] = {

            "Metadata": parsed_metadata,
            "Raw Data": {time_column:df[time_column].to_list(), signal_column:df[signal_column].to_list()}}#df.to_dict(orient="list")}
    tqdm.write(f"Found {len(txt_files)} .txt files.")
    if no_time_col:
        tqdm.write("Files missing time column:", no_time_col)
    if no_signal_col:
        tqdm.write("Files missing signal column:", no_signal_col)

    # Check against any existing dataset
    data_dict, unprocessed_keys = check_existing_jsons(data_dict, output_path)
    return {
        "data_dict": data_dict,
        "no_time_col": no_time_col,
        "no_signal_col": no_signal_col,
        "time_column": time_column,
        "signal_column": signal_column,
        "folder_path": folder_path,
        "output_path": output_path,
        "figures_path": figures_path,
        "unprocessed_samples": unprocessed_keys}

def parse_metadata_block(raw_text):
    """
    Parse chromatogram metadata string into a nested dictionary.
    """
    lines = raw_text.strip().split("\n")
    result = {}
    current_section = None

    for line in lines:
        if not line.strip():
            continue  # skip empty lines

        parts = line.split("\t")
        parts = [p.strip() for p in parts if p.strip()]

        if len(parts) == 1:
            # This is likely a section header like "Injection Information:"
            section = parts[0].rstrip(":")
            result[section] = {}
            current_section = section
        elif len(parts) == 2:
            key, value = parts
            if current_section:
                result[current_section][key] = value
            else:
                result[key] = value
        else:
            # Unhandled line structure
            tqdm.write("Skipping malformed line:", line)
            print("Skipping malformed line:", line())

    # return result

def create_output_folders(folder_path):
    """
    Creates a 'chromatoPy output' folder inside the given folder_path.
    If it already exists, deletes it and recreates it.
    Also creates a nested 'Figures' subfolder.

    Returns
    -------
    output_path : str
        Path to 'chromatoPy output' folder.
    figures_path : str
        Path to 'chromatoPy output/Figures' folder.
    """
    output_path = os.path.join(folder_path, "chromatoPy output")
    figures_path = os.path.join(output_path, "Figures")

    os.makedirs(output_path, exist_ok=True)
    os.makedirs(figures_path, exist_ok=True)

    return output_path, figures_path

def check_existing_jsons(data_dict, output_path):
    """
    Merge any new samples in data_dict into an existing JSON on disk.
    Returns: (merged_dict, unprocessed_keys)
      - merged_dict: either the existing JSON plus new samples, or your original data_dict
      - unprocessed_keys: list of sample-names that came in fresh
    """
    existing = load_json(output_path)
    # no file on disk → nothing processed yet
    if existing is None:
        unprocessed = list(data_dict["Samples"].keys())
        return data_dict, unprocessed

    # else: there was an existing JSON
    before = set(existing["Samples"].keys())
    # add any brand-new samples
    for name, sample in data_dict["Samples"].items():
        if name not in existing["Samples"]:
            existing["Samples"][name] = sample

    # figure out which ones we just added
    after = set(existing["Samples"].keys())
    unprocessed = list(after - before)

    return existing, unprocessed

def load_json(output_path, list_samples=False, list_processed=False):
    """
    Try to load FID_output.json from output_path.
    If it doesn’t exist, return None.
    Otherwise return the dict, rebuilding any Raw Data dicts into DataFrames.
    """
    js_file = os.path.join(output_path, "FID_output.json")
    if not os.path.exists(js_file):
        return None

    with open(js_file, "r") as f:
        data = json.load(f)

    # rebuild Raw Data dicts into DataFrames
    for sample in data.get("Samples", {}).values():
        raw = sample.get("Raw Data")
        if isinstance(raw, dict):
            sample["Raw Data"] = pd.DataFrame(raw)
    if list_samples:
        for key in data['Samples'].keys():
            print(key)
    if list_processed:
        key = []
        for x in data['Samples'].keys():
            if 'Processed Data' in data['Samples'][x].keys():
                key.append(x)
        print(key)
    return data




# ─── Standard Library ───────────────────────────────────────────────────────────
import os
import re
import sys
import shutil

# ─── Third-Party Libraries ─────────────────────────────────────────────────────
import numpy as np
import pandas as pd
import matplotlib
# matplotlib.use('Qt5Agg')
from matplotlib.widgets import TextBox, Button
import matplotlib.pyplot as plt
from scipy.signal import find_peaks
from tqdm import tqdm
import json

# ─── PyQt5 GUI Toolkit ─────────────────────────────────────────────────────────
from PyQt5.QtCore    import Qt
from PyQt5.QtWidgets import (
    QApplication,
    QDialog,
    QLineEdit,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QMessageBox)

# ─── Peak Integration ─────────────────────────────────────────────────────────
from .FID_Integration_functions import run_peak_integrator
from .FID_General import get_cluster_labels
from .manual_peak_integration import run_peak_integrator_manual
from .import_data import import_data

def integration(
        # Peak selection
        peak_labels=False, selection_method="nearest",
        # Data categorization
        categorized=None, 
        # Peak deconvolution 
        gaussian_fit_mode='single',  manual_peak_integration=False,
        # Peak integration parameters
        peak_neighborhood_n=3, smoothing_window=5, 
        smoothing_factor=3, gaus_iterations=1000, maximum_peak_amplitude=None, 
        peak_boundary_derivative_sensitivity=0.001, peak_prominence=0.01):
    """
    Main integration function for processing chromatographic samples.

    Parameters:
    - categorized: Dictionary of pre-categorized data. If None, raw data is imported.
    - selection_method: 'click' or 'nearest' for peak selection.
    - gaussian_fit_mode: 'single', 'multi', or 'both' for Gaussian fitting strategy.
    - peak_neighborhood_n: Maximum number of peaks in a neighborhood.
    - smoothing_window: Savitzky-Golay filter window size.
    - smoothing_factor: Savitzky-Golay polynomial order.
    - gaus_iterations: Max iterations for curve fitting.
    - maximum_peak_amplitude: Optional peak amplitude cap.
    - peak_boundary_derivative_sensitivity: Derivative threshold for boundary detection.
    - peak_prominence: Prominence threshold for peak finding.
    - peak_labels: If True, load peak label config from 'peak_labels.json'
    """
    
    # Handle predefined peak labels
    if peak_labels and manual_peak_integration:
        json_path = os.path.join(os.path.dirname(__file__), "peak_labels.json")
        if os.path.exists(json_path):
            with open(json_path, 'r') as f:
                peak_labels_data = json.load(f)
        else:
            raise FileNotFoundError(f"Expected peak_labels.json in {json_path}")
    else: peak_labels_data=None
    
    # Handle pre-categorized Data
    if categorized is not None:
        tqdm.write("Pre-categorized data.")
        result = import_data()
        unprocessed = result["unprocessed_samples"]
        if not unprocessed:
            return
        
        data = result['data_dict']
        time_column = result["time_column"]
        signal_column = result["signal_column"]
        folder_path = result["folder_path"]
        output_path = result["output_path"]
        figures_path = result["figures_path"]
        cluster_labels = get_cluster_labels(categorized)
        for i in cluster_labels:
            cluster_subset = {
                'Samples': {
                    key: value for key, value in data['Samples'].items()
                    if value.get('cluster') == i},
                'Integration Metadata': {}}
            FID_integration_backend(
                cluster_subset, time_column, signal_column,
                folder_path, output_path, figures_path,
                selection_method, gaussian_fit_mode,
                peak_neighborhood_n, smoothing_window,
                smoothing_factor, gaus_iterations,
                maximum_peak_amplitude, peak_boundary_derivative_sensitivity,
                peak_prominence,
                manual = manual_peak_integration,
                peak_labels=peak_labels_data)
    else:

        result = import_data()
        unprocessed = result["unprocessed_samples"]
        if not unprocessed:
            return
        data = result['data_dict']
        subset = {"Samples": {k: data["Samples"][k] for k in unprocessed},
          "Integration Metadata": {}}
        time_column = result["time_column"]
        signal_column = result["signal_column"]
        folder_path = result["folder_path"]
        output_path = result["output_path"]
        figures_path = result["figures_path"]
        FID_integration_backend(
            subset, time_column, signal_column,
            folder_path, output_path, figures_path,
            selection_method, gaussian_fit_mode,
            peak_neighborhood_n, smoothing_window, 
            smoothing_factor, gaus_iterations,
            maximum_peak_amplitude, peak_boundary_derivative_sensitivity,
            peak_prominence, manual = manual_peak_integration,
            peak_labels=peak_labels_data)
    
def print_no_samples_to_process():
        tqdm.write("All samples in this directory have been processed.")
        tqdm.write("To re-process sample entries, delete the samples with:")
        tqdm.write("    chromatopy.FID.delete_samples()")
        tqdm.write("Then rerun:")
        tqdm.write("    chromatopy.FID.integration()")
    
def FID_integration_backend(data, time_column, signal_column, folder_path, 
                            output_path, figures_path, sm, gaussian_fit_mode,
                            peak_neighborhood_n=3, smoothing_window=35, smoothing_factor=3, 
                            gaus_iterations=4000, maximum_peak_amplitude=None, 
                            peak_boundary_derivative_sensitivity=0.01, peak_prominence=1, 
                            manual=False, peak_labels=None):
    
    # Get unprocessed samples only
    unprocessed_keys = [k for k in data["Samples"].keys() if 'Processed Data' not in data["Samples"][k].keys()]
    
    # Identify peak locations
    if manual and peak_labels is not None:
        tqdm.write("Using stored peak labels for manual integration.")
        data['Integration Metadata'] = {
            "peak dictionary": peak_labels["Peak Labels"],
            "x limits": peak_labels["x limits"],
            "time_column": time_column,
            "signal_column": signal_column}
    else:
        tqdm.write("Click the location of peaks and enter the chain length of interest (e.g., C22).\nUse 'shift+delete' to remove the last peak.\n'Select 'Finished' once satisfied.")
        # app = QApplication.instance() or QApplication(sys.argv)
        # first_key = unprocessed_keys[0]
        # time = data['Samples'][first_key]['Raw Data'][time_column]
        # signal = data['Samples'][first_key]['Raw Data'][signal_column]
    
        # if sm == "nearest":
        #     peak_positions, _ = find_peaks(signal)
        # elif sm == "click":
        #     peak_positions = None
    
        # peak_identifier = FID_Peak_ID(x=time, y=signal, selection_method=sm, peak_positions=peak_positions)
        # app.exec_()
        
        app = QApplication.instance()
        owns_app = False
        if app is None:
            app = QApplication(sys.argv)
            owns_app = True  # we created it; safe to run/quit
        
        first_key = unprocessed_keys[0]
        time = data['Samples'][first_key]['Raw Data'][time_column]
        signal = data['Samples'][first_key]['Raw Data'][signal_column]
        
        if sm == "nearest":
            peak_positions, _ = find_peaks(signal)
        elif sm == "click":
            peak_positions = None
        
        peak_identifier = FID_Peak_ID(
            x=time,
            y=signal,
            selection_method=sm,
            peak_positions=peak_positions,
            owns_app=owns_app)
        
        if owns_app:
            # Script/terminal case: we own the event loop
            app.exec_()
        else:
            # Spyder/Jupyter: event loop already integrated. Just wait until the figure closes.
            # (Works with either `%matplotlib widget` (recommended) or `%matplotlib qt`)
            while plt.fignum_exists(peak_identifier.fig.number) and not getattr(peak_identifier, "result", None):
                plt.pause(0.1)
        if peak_identifier.result is None:
            # User closed the window without pressing Finish → cancel/abort
            import tqdm as tqdmmod
            tqdmmod.write("Integration cancelled: window closed without pressing 'Finish'.")
            raise SystemExit
    
        data['Integration Metadata'] = {
            "peak dictionary": peak_identifier.result,
            "time_column": time_column,
            "signal_column": signal_column}
    
    for key in tqdm(unprocessed_keys, desc="Integrating samples", unit="sample", mininterval=0, maxinterval=0):
        if "Integratoin Result" in data['Samples'][key].keys():
            tqdm.write(f"{key} already processed")
            continue
    
        if manual:
            run_peak_integrator_manual(data, key, gi=gaus_iterations,
                                       pk_sns=peak_boundary_derivative_sensitivity,
                                       smoothing_params=[smoothing_window, smoothing_factor],
                                       max_peaks_for_neighborhood=peak_neighborhood_n,
                                       fp=figures_path,
                                       gaussian_fit_mode=gaussian_fit_mode)
        else:
            run_peak_integrator(data, key, gi=gaus_iterations,
                                pk_sns=peak_boundary_derivative_sensitivity,
                                smoothing_params=[smoothing_window, smoothing_factor],
                                max_peaks_for_neighborhood=peak_neighborhood_n,
                                fp=figures_path,
                                gaussian_fit_mode=gaussian_fit_mode)
        data = round_dict_floats(data)
        existing_data = load_json(output_path)
        if existing_data:
            for sample_name, sample_data in data["Samples"].items():
                existing_data["Samples"][sample_name] = sample_data
        else:
            existing_data = data
    
        save_json({'data_dict':existing_data}, output_path)
        output_csv(existing_data, output_path)
    # return data
    
def output_csv(data, output_directory):
    """
    Generate two CSV files:
    1. output_peak_areas.csv: contains median peak areas per sample.
    2. output_retention_times.csv: contains retention times per sample.
    
    Parameters
    ----------
    data : dict
        Data dictionary containing processed chromatographic data.
    output_directory : str
        Directory to save the output CSVs.
    
    Returns
    -------
    df_areas : pandas.DataFrame
        DataFrame of peak area medians per sample.
    df_ret_times : pandas.DataFrame
        DataFrame of retention times per sample.
    """
    # Collect all unique peak labels across samples
    all_peaks = set()
    for sample in data['Samples'].values():
        processed = sample.get("Processed Data", {})
        all_peaks.update(processed.keys())
    
    all_peaks = sorted(all_peaks)
    
    peak_area_rows = []
    retention_time_rows = []
    
    for sample_name, sample in data['Samples'].items():
        row_area = {"Lab ID": sample_name}
        row_ret = {"Lab ID": sample_name}
        processed = sample.get("Processed Data", {})
        for peak in all_peaks:
            peak_data = processed.get(peak, None)
            if peak_data and isinstance(peak_data, dict):
                row_area[peak] = peak_data.get("Peak Area - median", np.nan)
                row_ret[peak] = peak_data.get("Retention Time", np.nan)
            else:
                row_area[peak] = np.nan
                row_ret[peak] = np.nan
        peak_area_rows.append(row_area)
        retention_time_rows.append(row_ret)
    
    # Convert to DataFrames
    df_areas = pd.DataFrame(peak_area_rows)
    df_ret_times = pd.DataFrame(retention_time_rows)
    
    # Save as CSVs
    os.makedirs(output_directory, exist_ok=True)
    area_path = os.path.join(output_directory, "output_peak_areas.csv")
    rt_path = os.path.join(output_directory, "output_retention_times.csv")
    df_areas.to_csv(area_path, index=False)
    df_ret_times.to_csv(rt_path, index=False)

def convert_dataframes_to_dicts(obj):
    """
    Walks any nested structure of dicts/lists and converts
    any pandas.DataFrame it finds into a plain dict of lists.
    """
    if isinstance(obj, pd.DataFrame):
        return obj.to_dict(orient="list")
    elif isinstance(obj, dict):
        return {k: convert_dataframes_to_dicts(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [convert_dataframes_to_dicts(v) for v in obj]
    else:
        return obj

def save_json(container, output_path):
    # “container” is the dict you get back from import_data()
    data_dict = container["data_dict"]

    # convert _all_ DataFrames under data_dict → dicts of lists
    cleanable = convert_dataframes_to_dicts(data_dict)

    js_file = os.path.join(output_path, "FID_output.json")
    os.makedirs(os.path.dirname(js_file), exist_ok=True)
    with open(js_file, "w") as f:
        json.dump(clean_for_json(cleanable), f, indent=4)

def load_json(output_path, list_samples=False, list_processed=False):
    """
    Try to load FID_output.json from output_path.
    If it doesn’t exist, return None.
    Otherwise return the dict, rebuilding any Raw Data dicts into DataFrames.
    """
    js_file = os.path.join(output_path, "FID_output.json")
    if not os.path.exists(js_file):
        return None

    with open(js_file, "r") as f:
        data = json.load(f)

    # rebuild Raw Data dicts into DataFrames
    for sample in data.get("Samples", {}).values():
        raw = sample.get("Raw Data")
        if isinstance(raw, dict):
            sample["Raw Data"] = pd.DataFrame(raw)
    if list_samples:
        for key in data['Samples'].keys():
            print(key)
    if list_processed:
        key = []
        for x in data['Samples'].keys():
            if 'Processed Data' in data['Samples'][x].keys():
                key.append(x)
        print(key)
    return data


# def create_output_folders(folder_path):
#     """
#     Creates a 'chromatoPy output' folder inside the given folder_path.
#     If it already exists, deletes it and recreates it.
#     Also creates a nested 'Figures' subfolder.

#     Returns
#     -------
#     output_path : str
#         Path to 'chromatoPy output' folder.
#     figures_path : str
#         Path to 'chromatoPy output/Figures' folder.
#     """
#     output_path = os.path.join(folder_path, "chromatoPy output")
#     figures_path = os.path.join(output_path, "Figures")
    
#     os.makedirs(output_path, exist_ok=True)
#     os.makedirs(figures_path, exist_ok=True)

#     return output_path, figures_path

def clean_for_json(obj):
    if isinstance(obj, (np.integer, np.floating)):
        return obj.item()
    elif isinstance(obj, (np.ndarray, pd.Series, list, tuple)):
        return [clean_for_json(el) for el in obj]
    elif isinstance(obj, pd.DataFrame): 
        return obj.to_dict(orient="list")
    elif isinstance(obj, dict):
        return {str(k): clean_for_json(v) for k, v in obj.items()}
    else:
        try:
            json.dumps(obj)
            return obj
        except (TypeError, OverflowError):
            return str(obj)
        
# def parse_metadata_block(raw_text):
#     """
#     Parse chromatogram metadata string into a nested dictionary.
#     """
#     lines = raw_text.strip().split("\n")
#     result = {}
#     current_section = None

#     for line in lines:
#         if not line.strip():
#             continue  # skip empty lines

#         parts = line.split("\t")
#         parts = [p.strip() for p in parts if p.strip()]

#         if len(parts) == 1:
#             # This is likely a section header like "Injection Information:"
#             section = parts[0].rstrip(":")
#             result[section] = {}
#             current_section = section
#         elif len(parts) == 2:
#             key, value = parts
#             if current_section:
#                 result[current_section][key] = value
#             else:
#                 result[key] = value
#         else:
#             # Unhandled line structure
#             #tqdm.write("Skipping malformed line:", line)

#     return result
 
def merge_existing_jsons(data, output_path):
    """
    Function Identifies previously processed data (.json files) in working
    directory, selects samples from this previous dataset that are not in
    the raw .txt files, and adds these samples to the new dictionary.
    --- Requires heavy modificatoin to properly retain model information.

    Parameters
    ----------
    data : dict
        Imported data from .txt file. Contains raw data
    output_path : str
        Location of output directory.

    Returns
    -------
    data : dict
        Dataset.
    """
    # output_path, figures_path = create_output_folders(folder_path)
    existing_data = load_json(output_path)
    new_samples = {}
    if existing_data is not None:
        for sample_name, sample in data["Samples"].items():
            if sample_name not in existing_data["Samples"]:
                existing_data["Samples"][sample_name] = sample
        data = existing_data
    save_json(data, output_path)
    return data


# def check_existing_jsons(data_dict, output_path):
#     """
#     Merge any new samples in data_dict into an existing JSON on disk.
#     Returns: (merged_dict, unprocessed_keys)
#       - merged_dict: either the existing JSON plus new samples, or your original data_dict
#       - unprocessed_keys: list of sample-names that came in fresh
#     """
#     existing = load_json(output_path)
#     # no file on disk → nothing processed yet
#     if existing is None:
#         unprocessed = list(data_dict["Samples"].keys())
#         return data_dict, unprocessed

#     # else: there was an existing JSON
#     before = set(existing["Samples"].keys())
#     # add any brand-new samples
#     for name, sample in data_dict["Samples"].items():
#         if name not in existing["Samples"]:
#             existing["Samples"][name] = sample

#     # figure out which ones we just added
#     after = set(existing["Samples"].keys())
#     unprocessed = list(after - before)

#     return existing, unprocessed


# class FID_Peak_ID:
#     def __init__(self, x, y, selection_method, peak_positions=None, owns_app=False):
#         self._owns_app = owns_app
#         self.result=None
#         self.finished=False
#         self.closed_without_finish = False
#         self.x = pd.Series(x)
#         self.y = pd.Series(y)
#         self.selection_method = selection_method
#         self.lines = []
#         self.labels = []
#         self.positions = set()
#         self.peak_dict = {}
#         self.peak_order = []

#         if peak_positions is None:
#             self.peak_positions = []
#         else:
#             self.peak_positions = list(peak_positions)

#         # Create figure and axes
#         self.fig, self.ax = plt.subplots(figsize=(10, 5))
#         self.ax.plot(self.x, self.y)

#         # TextBoxes for axis limits
#         self.textbox_minx_ax = self.fig.add_axes([0.15, 0.02, 0.1, 0.04])
#         self.textbox_maxx_ax = self.fig.add_axes([0.3, 0.02, 0.1, 0.04])
#         self.textbox_miny_ax = self.fig.add_axes([0.45, 0.02, 0.1, 0.04])
#         self.textbox_maxy_ax = self.fig.add_axes([0.6, 0.02, 0.1, 0.04])
#         self.textbox_minx = TextBox(self.textbox_minx_ax, 'X0')
#         self.textbox_maxx = TextBox(self.textbox_maxx_ax, 'X1')
#         self.textbox_miny = TextBox(self.textbox_miny_ax, 'Y0')
#         self.textbox_maxy = TextBox(self.textbox_maxy_ax, 'Y1')

#         # Initialize limits
#         xmin, xmax = self.ax.get_xlim()
#         ymin, ymax = self.ax.get_ylim()
#         self.textbox_minx.set_val(str(round(xmin, 1)))
#         self.textbox_maxx.set_val(str(round(xmax, 1)))
#         self.textbox_miny.set_val(str(round(ymin, 1)))
#         self.textbox_maxy.set_val(str(round(ymax, 1)))

#         # Connect axis limit updates
#         for box in [self.textbox_minx, self.textbox_maxx, self.textbox_miny, self.textbox_maxy]:
#             box.on_submit(self.update_limits)

#         # # Finish button
#         # self.button_ax = self.fig.add_axes([0.85, 0.02, 0.1, 0.04])
#         # self.finish_button = Button(self.button_ax, 'Finished')
#         # self.finish_button.on_clicked(self.finish)
#         self.fig.canvas.mpl_connect("close_event", self._on_close)
        
#         # def finish(self, event):
#         #     # Store results for the caller
#         #     self.result = dict(self.peak_dict)
#         #     # Close the plot
#         #     plt.close(self.fig)
        
#         #     # Only quit the app if THIS code created/owns it
#         #     if self._owns_app:
#         #         app = QApplication.instance()
#         #         if app is not None:
#         #             app.quit()
#         def finish(self, event):
#             self.result = dict(self.peak_dict)  # whatever you collect
#             self.finished = True
#             import matplotlib.pyplot as plt
#             plt.close(self.fig)
#             if self._owns_app:
#                 from PyQt5.QtWidgets import QApplication
#                 app = QApplication.instance()
#                 if app is not None:
#                     app.quit()
#         def _on_close(self, event):
#             if not self.finished:
#                 self.closed_without_finish = True
                
#         # Connect events
#         self.fig.canvas.mpl_connect('button_press_event', self.on_click)
#         self.fig.canvas.mpl_connect('key_press_event', self.on_key)

#         plt.show()

#         # Focus - permits seeing key events
#         self.fig.canvas.setFocusPolicy(Qt.StrongFocus)
#         self.fig.canvas.setFocus()
#         def _on_close(self, event):
        
#     def on_click(self, event):
#             if event.inaxes != self.ax:
#                 return
#             # x_click = round(event.xdata, 5)
#             # if x_click in self.positions:
#             #     return

#             if self.selection_method == "nearest" and self.peak_positions:
#                 raw_x = event.xdata
#                 # find the peak position closest to where they clicked
#                 # x_click = min(self.x[self.peak_positions], key=lambda xp: abs(xp - raw_x))
#                 peak_times = self.x.iloc[self.peak_positions].to_numpy()
#                 x_click = min(peak_times, key=lambda t: abs(t - raw_x))
#             else:
#                 x_click = round(event.xdata, 5)

#             if x_click in self.positions:
#                 return

#             # 1) draw the line immediately (so user sees it)
#             line = self.ax.axvline(x_click, color='red', linestyle='--', alpha=0.7)
#             self.lines.append(line)

#             # 2) ask for the label via our Qt dialog
#             prompt = f"Label for peak at x = {x_click:.2f}"
#             dlg = LabelDialog(prompt=prompt, initial="peak", parent=self.fig.canvas)
#             if dlg.exec_() == QDialog.Accepted:
#                 text = dlg.value()
#                 # duplicate‐name check
#                 if text in self.peak_dict:
#                     QMessageBox.warning(
#                         self.fig.canvas, "Duplicate label",
#                         f"'{text}' already exists—please pick another name."
#                     )
#                     # undo that line
#                     self.lines.pop().remove()
#                     return

#                 # 3) record & draw the annotation
#                 self.positions.add(x_click)
#                 self.peak_order.append(text)
#                 self.peak_dict[text] = x_click

#                 y_top = self.ax.get_ylim()[1]
#                 txt = self.ax.text(
#                     x_click + 0.05, y_top * 0.95, text, #rotation=90,
#                     verticalalignment='top', horizontalalignment='left',
#                     color='red', fontsize=9,
#                     bbox=dict(facecolor='white', alpha=0.5)
#                 )
#                 self.labels.append(txt)
#                 self.fig.canvas.draw_idle()

#             else:
#                 # user cancelled → remove that line
#                 self.lines.pop().remove()
#                 self.fig.canvas.draw_idle()

#     def on_key(self, event):
#         qt_ev = getattr(event, "guiEvent", None)
#         if not qt_ev:
#             return

#         keycode = qt_ev.key()
#         # Qt.Key_Backspace = 16777219, Qt.Key_Delete = 16777223
#         if (qt_ev.modifiers() & Qt.ShiftModifier) and keycode in (Qt.Key_Backspace, Qt.Key_Delete):
#             if not self.peak_order:
#                 return

#             # 1) remove last vertical line
#             line = self.lines.pop()
#             line.remove()

#             # 2) remove last text annotation
#             txt = self.labels.pop()
#             txt.remove()

#             # 3) clean up bookkeeping
#             last_label = self.peak_order.pop()
#             x_removed = self.peak_dict.pop(last_label)
#             self.positions.discard(x_removed)

#             # 4) redraw
#             self.fig.canvas.draw_idle()

#     def update_limits(self, _):
#         # Update axis limits from TextBox values
#         try:
#             xmin = float(self.textbox_minx.text)
#             xmax = float(self.textbox_maxx.text)
#             ymin = float(self.textbox_miny.text)
#             ymax = float(self.textbox_maxy.text)
#             self.ax.set_xlim(xmin, xmax)
#             self.ax.set_ylim(ymin, ymax)
#             self.fig.canvas.draw_idle()
#         except ValueError:
#             print("oh no")
#             tqdm.write('Invalid axis limits entered.')
#         finally:
#             # force-release any mouse grab so next click is clean
#             try:
#                 self.fig.canvas.release_mouse(self.ax)
#             except Exception:
#                 pass

#     # def finish(self, event):
#     #     # Store the peaks dict so callers can grab it
#     #     self.result = dict(self.peak_dict)
#     #     # Close the plot
#     #     plt.close(self.fig)
#     #     # Quit the Qt event loop so exec_() returns
#     #     QApplication.instance().quit()
#     def finish(self, event):
#         # Store results and mark as finished
#         self.result = dict(self.peak_dict)
#         self.finished = True
    
#         # Close the figure
#         plt.close(self.fig)
    
#         # Only quit the Qt loop if WE created it
#         if self._owns_app:
#             app = QApplication.instance()
#             if app is not None:
#                 app.quit()
    
class FID_Peak_ID:
    def __init__(self, x, y, selection_method, peak_positions=None, owns_app=False):
        self._owns_app = owns_app
        self.result = None
        self.finished = False
        self.closed_without_finish = False

        self.x = pd.Series(x)
        self.y = pd.Series(y)
        self.selection_method = selection_method
        self.lines = []
        self.labels = []
        self.positions = set()
        self.peak_dict = {}
        self.peak_order = []
        self.peak_positions = list(peak_positions) if peak_positions is not None else []

        # Figure & plot
        self.fig, self.ax = plt.subplots(figsize=(10, 5))
        self.ax.plot(self.x, self.y)
        self.fig.canvas.mpl_connect("close_event", self._on_close)

        # Axis limit text boxes
        self.textbox_minx_ax = self.fig.add_axes([0.15, 0.02, 0.1, 0.04])
        self.textbox_maxx_ax = self.fig.add_axes([0.30, 0.02, 0.1, 0.04])
        self.textbox_miny_ax = self.fig.add_axes([0.45, 0.02, 0.1, 0.04])
        self.textbox_maxy_ax = self.fig.add_axes([0.60, 0.02, 0.1, 0.04])
        self.textbox_minx = TextBox(self.textbox_minx_ax, 'X0')
        self.textbox_maxx = TextBox(self.textbox_maxx_ax, 'X1')
        self.textbox_miny = TextBox(self.textbox_miny_ax, 'Y0')
        self.textbox_maxy = TextBox(self.textbox_maxy_ax, 'Y1')

        xmin, xmax = self.ax.get_xlim()
        ymin, ymax = self.ax.get_ylim()
        self.textbox_minx.set_val(str(round(xmin, 1)))
        self.textbox_maxx.set_val(str(round(xmax, 1)))
        self.textbox_miny.set_val(str(round(ymin, 1)))
        self.textbox_maxy.set_val(str(round(ymax, 1)))

        for box in [self.textbox_minx, self.textbox_maxx, self.textbox_miny, self.textbox_maxy]:
            box.on_submit(self.update_limits)

        # Finished button (UNCOMMENTED & hooked up)
        self.button_ax = self.fig.add_axes([0.85, 0.02, 0.1, 0.04])
        self.finish_button = Button(self.button_ax, 'Finished')
        self.finish_button.on_clicked(self.finish)

        # Events
        self.fig.canvas.mpl_connect('button_press_event', self.on_click)
        self.fig.canvas.mpl_connect('key_press_event', self.on_key)

        plt.show()

        # Focus for key events
        self.fig.canvas.setFocusPolicy(Qt.StrongFocus)
        self.fig.canvas.setFocus()

    # ---------- event handlers (class methods, not nested) ----------
    def _on_close(self, event):
        if not self.finished:
            self.closed_without_finish = True

    def finish(self, event):
        self.result = dict(self.peak_dict)
        self.finished = True
        plt.close(self.fig)
        if self._owns_app:
            app = QApplication.instance()
            if app is not None:
                app.quit()

    def on_click(self, event):
        if event.inaxes != self.ax:
            return

        if self.selection_method == "nearest" and self.peak_positions:
            raw_x = event.xdata
            peak_times = self.x.iloc[self.peak_positions].to_numpy()
            x_click = min(peak_times, key=lambda t: abs(t - raw_x))
        else:
            x_click = round(event.xdata, 5)

        if x_click in self.positions:
            return

        line = self.ax.axvline(x_click, color='red', linestyle='--', alpha=0.7)
        self.lines.append(line)

        prompt = f"Label for peak at x = {x_click:.2f}"
        dlg = LabelDialog(prompt=prompt, initial="peak", parent=self.fig.canvas)
        if dlg.exec_() == QDialog.Accepted:
            text = dlg.value()
            if text in self.peak_dict:
                QMessageBox.warning(self.fig.canvas, "Duplicate label",
                                    f"'{text}' already exists—please pick another name.")
                self.lines.pop().remove()
                return

            self.positions.add(x_click)
            self.peak_order.append(text)
            self.peak_dict[text] = x_click

            y_top = self.ax.get_ylim()[1]
            txt = self.ax.text(
                x_click + 0.05, y_top * 0.95, text,
                va='top', ha='left', color='red', fontsize=9,
                bbox=dict(facecolor='white', alpha=0.5)
            )
            self.labels.append(txt)
            self.fig.canvas.draw_idle()
        else:
            self.lines.pop().remove()
            self.fig.canvas.draw_idle()

    def on_key(self, event):
        qt_ev = getattr(event, "guiEvent", None)
        if not qt_ev:
            return
        keycode = qt_ev.key()
        if (qt_ev.modifiers() & Qt.ShiftModifier) and keycode in (Qt.Key_Backspace, Qt.Key_Delete):
            if not self.peak_order:
                return
            self.lines.pop().remove()
            self.labels.pop().remove()
            last_label = self.peak_order.pop()
            x_removed = self.peak_dict.pop(last_label)
            self.positions.discard(x_removed)
            self.fig.canvas.draw_idle()

    def update_limits(self, _):
        try:
            xmin = float(self.textbox_minx.text)
            xmax = float(self.textbox_maxx.text)
            ymin = float(self.textbox_miny.text)
            ymax = float(self.textbox_maxy.text)
            self.ax.set_xlim(xmin, xmax)
            self.ax.set_ylim(ymin, ymax)
            self.fig.canvas.draw_idle()
        except ValueError:
            print('Invalid axis limits entered.')
        finally:
            try:
                self.fig.canvas.release_mouse(self.ax)
            except Exception:
                pass 

class LabelDialog(QDialog):
    def __init__(self, prompt="Enter peak label:", initial="peak", parent=None):
        super().__init__(parent)
        self.setWindowTitle("Peak Label")
        # Prompt text
        lbl = QLabel(prompt, self)
        # The line‐edit
        self.edit = QLineEdit(self)
        self.edit.setText(initial)
        self.edit.selectAll()           # select all so typing replaces
        # OK button
        ok = QPushButton("OK", self)
        ok.clicked.connect(self.accept)
        # Layout
        layout = QVBoxLayout(self)
        layout.addWidget(lbl)
        layout.addWidget(self.edit)
        layout.addWidget(ok)
        self.setLayout(layout)

    def value(self):
        return self.edit.text().strip()

def round_dict_floats(data, default_decimals=4):
    """
    Recursively round all floats in a nested dictionary or list structure to a fixed number of decimals.
    Handles DataFrames by rounding numeric columns.
    Does not round stringified DataFrame representations.

    Parameters:
    - data: dict or list
    - default_decimals: int, number of decimal places to round to
    """
    def process_df(df):
        for col in df.columns:
            if pd.api.types.is_numeric_dtype(df[col]):
                df[col] = df[col].round(default_decimals)
        return df

    def recursive_round(obj):
        if isinstance(obj, dict):
            new_dict = {}
            for k, v in obj.items():
                if k == "Raw Data" and isinstance(v, pd.DataFrame):
                    new_dict[k] = process_df(v.copy())
                else:
                    new_dict[k] = recursive_round(v)
            return new_dict
        elif isinstance(obj, list):
            return [recursive_round(v) for v in obj]
        elif isinstance(obj, pd.DataFrame):
            return process_df(obj.copy())
        elif isinstance(obj, (float, np.floating)):
            return round(obj, default_decimals)
        else:
            return obj

    return recursive_round(data)


import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import find_peaks, savgol_filter
from scipy.integrate import simpson
from scipy.optimize import curve_fit
import math
from scipy.integrate import simpson
from pybaselines import Baseline
import warnings 
import pandas as pd
from tqdm import tqdm
from scipy.special import erf
import matplotlib.pyplot as plt

# Functions
def baseline( x, y, deg=5, max_it=1000, tol=1e-4):
    original_y = y.copy()
    order = deg + 1
    coeffs = np.ones(order)
    cond = math.pow(abs(y).max(), 1.0 / order)
    x = np.linspace(0.0, cond, y.size)  # Ensure this generates the expected range
    base = y.copy()
    vander = np.vander(x, order)  # Could potentially generate huge matrix if misconfigured
    vander_pinv = np.linalg.pinv(vander)
    for _ in range(max_it):
        coeffs_new = np.dot(vander_pinv, y)
        if np.linalg.norm(coeffs_new - coeffs) / np.linalg.norm(coeffs) < tol:
            break
        coeffs = coeffs_new
        base = np.dot(vander, coeffs)
        y = np.minimum(y, base)

    # Calculate maximum peak amplitude (3 x baseline amplitude)
    baseline_fitter = Baseline(x)
    fit, params_mask = baseline_fitter.std_distribution(y, 45)#, smooth_half_window=10)
    mask = params_mask['mask'] #  Mask for regions of signal without peaks
    min_peak_amp = (np.std(y[mask]))*2*3 # 2 sigma times 3
    return base, min_peak_amp # return base

def find_valleys(y, peaks, peak_oi=None):
    valleys = []
    if peak_oi == None:
        for i in range(1, len(peaks)):
            valley_point = np.argmin(y[peaks[i - 1] : peaks[i]]) + peaks[i - 1]
            valleys.append(valley_point)
    else:
        poi = np.where(peaks == peak_oi)[0][0]
        valleys.append(np.argmin(y[peaks[poi - 1] : peaks[poi]]) + peaks[poi - 1])
        valleys.append(np.argmin(y[peaks[poi] : peaks[poi + 1]]) + peaks[poi])
    return valleys

# def smoother(y, param_0, param_1, mode = "interp"):# "constant"):
#     return savgol_filter(y, param_0, param_1, mode=mode)
def smoother(y, window_length, polyorder):
    from scipy.signal import savgol_filter

    if len(y) < 3:
        return y  # don't try to smooth tiny series

    # Adjust window_length to be <= len(y) and an odd integer
    window_length = min(window_length, len(y) - 1 if len(y) % 2 == 0 else len(y))
    if window_length % 2 == 0:
        window_length -= 1
    window_length = max(window_length, polyorder + 2 + (polyorder % 2))  # ensure still valid

    return savgol_filter(y, window_length=window_length, polyorder=polyorder)

def find_peak_neighborhood_boundaries(x, y_smooth, peaks, valleys, peak_idx, max_peaks, peak_properties, gi, smoothing_params, pk_sns):
    overlapping_peaks = []
    extended_boundaries = {}
    # Analyze each of the closest peaks
    for peak in peaks: #closest_peaks:
        peak_pos = np.where(peak == peaks)
        l_lim = peak_properties["left_bases"][peak_pos][0]
        r_lim = peak_properties["right_bases"][peak_pos][0]
        heights, means, stddevs = estimate_initial_gaussian_params(x[l_lim : r_lim + 1], y_smooth[l_lim : r_lim + 1], peak)
        height, mean, stddev = heights[0], means[0], stddevs[0]

        # Fit Gaussian and get best fit parameters
        try:
            popt, _ = curve_fit(individual_gaussian, x, y_smooth, p0=[height, mean, stddev], maxfev=gi)
        except RuntimeError:
            popt, _ = curve_fit(individual_gaussian, x, y_smooth, p0=[height, mean, stddev], maxfev=gi*100)
        # Extend Gaussian fit limits
        x_min, x_max = calculate_gaus_extension_limits(popt[1], popt[2], 0, factor=3)
        extended_x, extended_y = extrapolate_gaussian(x, popt[0], popt[1], popt[2], None, x_min - 2, x_max + 2)
        # Find the boundaries based on the derivative test
        peak_x_value = x[peak]
        n_peak_idx = np.argmin(np.abs(extended_x - peak_x_value))
        left_idx, right_idx = calculate_boundaries(extended_x, extended_y, n_peak_idx, smoothing_params, pk_sns)
        extended_boundaries[peak] = (extended_x[left_idx], extended_x[right_idx])

    # Determine the peak of interest boundaries
    poi_bounds = extended_boundaries.get(peak_idx, (None, None))

    # Check for overlaps and determine the neighborhood
    for peak, bounds in extended_boundaries.items():
        if peak < peak_idx and bounds[1] > poi_bounds[0]:  # Overlaps to the left
            overlapping_peaks.append(peak)
        elif peak > peak_idx and bounds[0] < poi_bounds[1]:  # Overlaps to the right
            overlapping_peaks.append(peak)

    # Calculate neighborhood boundaries based on the left-most and right-most overlapping peaks
    if overlapping_peaks:
        left_most_peak = min(overlapping_peaks, key=lambda p: extended_boundaries[p][0])
        right_most_peak = max(overlapping_peaks, key=lambda p: extended_boundaries[p][1])
        neighborhood_left_boundary = extended_boundaries[left_most_peak][0]
        neighborhood_right_boundary = extended_boundaries[right_most_peak][1]
    else:
        # Use the peak of interest's bounds if no other peaks are overlapping
        neighborhood_left_boundary = poi_bounds[0]
        neighborhood_right_boundary = poi_bounds[1]
    return neighborhood_left_boundary, neighborhood_right_boundary, overlapping_peaks


# Gaussian fitting
def calculate_gaus_extension_limits(cen, wid, decay, factor=3, max_tail_sigma=5):
    sigma_effective = wid * factor  # Adjust factor for tail thinness
    if decay <= 0:
        tail = sigma_effective * max_tail_sigma
    else:
        tail = min(1/decay, sigma_effective * max_tail_sigma)
    return cen - sigma_effective-tail, cen+sigma_effective+tail

def extrapolate_gaussian(x, amp, cen, wid, skew=None, x_min=None, x_max=None, step=0.0001):
    if x_min is None: x_min = cen - 3 * wid
    if x_max is None: x_max = cen + 3 * wid
    extended_x = np.arange(x_min, x_max, step)

    if skew is None:
        extended_y = individual_gaussian(extended_x, amp, cen, wid)
    else:
        extended_y = skewed_gaussian(extended_x, amp, cen, wid, skew)

    return extended_x, extended_y

def calculate_boundaries(x, y, ind_peak, smoothing_params, pk_sns):
    smooth_y = smoother(y, smoothing_params[0], smoothing_params[1])
    velocity, X1 = forward_derivative(x, smooth_y)
    velocity /= np.max(np.abs(velocity))
    if smoothing_params[0] > len(velocity):
        smoother_val = len(velocity)-1
    else: smoother_val = smoothing_params[0]
    smooth_velo = smoother(velocity, smoother_val, smoothing_params[1])
    dt = int(np.ceil(0.025 / np.mean(np.diff(x))))
    A = np.where(smooth_velo[: ind_peak - 3 * dt] < pk_sns)[0]  # 0.05)[0]
    B = np.where(smooth_velo[ind_peak + 3 * dt :] > -pk_sns)[0]  # -0.05)[0]
    if A.size > 0:
        A = A[-1] + 1
    else:
        A = 1
    if B.size > 0:
        B = B[0] + ind_peak + 3 * dt - 1
    else:
        B = len(x) - 1
    return A, B


def calculate_boundaries_acceleration(x, y, ind_peak, smoothing_params, pk_sns):
    smooth_y = smoother(y, smoothing_params[0], smoothing_params[1])
    velocity, _ = forward_derivative(x, smooth_y)
    acceleration, _ = forward_derivative(x[:-1], velocity)
    acceleration /= np.max(np.abs(acceleration))
    smoother_val = min(smoothing_params[0], len(acceleration) - 1)
    smooth_accel = smoother(acceleration, smoother_val, smoothing_params[1])
    left_zone = smooth_accel[:ind_peak]
    right_zone = smooth_accel[ind_peak:]
    if len(left_zone) > 0:
        A = np.argmax(left_zone)
    else:
        A = 1
    if len(right_zone) > 0:
        B = np.argmax(right_zone) + ind_peak
    else:
        B = len(x) - 1
    return A, B

def fit_gaussians(x_full, y_full, ind_peak, peaks, smoothing_params, pk_sns, gi, mode="both"):
    if mode not in {"single", "multi", "both"}:
        raise ValueError("mode must be 'single', 'multi', or 'both'")
    # figy = plt.figure()
    results = []
    
    # --- MULTI-GAUSSIAN ---
    if mode in {"multi", "both"}:
        result = _fit_multi_gaussian(x_full, y_full, ind_peak, peaks, smoothing_params, pk_sns, gi)
        if result is not None:
            best_x, best_fit_y, best_fit_params, best_fit_params_error, best_error, best_idx_interest = result
            results.append({
                "name": "multi",
                "x": best_x,
                "y": best_fit_y,
                "params": best_fit_params,
                "pcov": best_fit_params_error,
                "error": best_error,
                "idx_interest": best_idx_interest,
                "multi_flag": True})

    # --- SINGLE-GAUSSIAN ---
    if mode in {"single", "both"}:
        # print("debug 1.1")
        result = _fit_single_gaussian(x_full, y_full, ind_peak, smoothing_params, pk_sns, gi, current_best_error=float("inf"))
        # print("debug 1.2")
        if result is not None:
            best_x, best_fit_y, best_fit_params, best_fit_params_error, best_error = result
            results.append({
                "name": "single",
                "x": best_x,
                "y": best_fit_y,
                "params": best_fit_params,
                "pcov": best_fit_params_error,
                "error": best_error,
                "multi_flag": False,
                "idx_interest": None})

    # --- ASYMMETRIC MODEL (always run) ---
    
    # print("debug 1.3")
    result = _fit_asymmetric_gaussian(x_full, y_full, ind_peak, smoothing_params, pk_sns, gi, current_best_error=float("inf"))
    # print("debug 1.4")
    if result is not None:
        best_x, best_fit_y, best_fit_params, best_fit_params_error, best_error = result
        results.append({
            "name": "asymmetric",
            "x": best_x,
            "y": best_fit_y,
            "params": best_fit_params,
            "pcov": best_fit_params_error,
            "error": best_error,
            "multi_flag": False,
            "idx_interest": None})
    if not results:
        raise RuntimeError(f"No valid fit found for peak at index {ind_peak}")

    best_result = min(results, key=lambda r: r["error"])
    # print("debug 1.5")
    # --- Process best fit output ---
    best_x = best_result["x"]
    best_fit_y = best_result["y"]
    best_fit_params = best_result["params"]
    best_fit_params_error = best_result["pcov"]
    best_idx_interest = best_result.get("idx_interest", None)
    multi_gauss_flag = best_result["multi_flag"]
    model_used = best_result["name"]

    # --- Extend fit + calculate area ---
    if multi_gauss_flag:
        amp, cen, wid = best_fit_params[best_idx_interest * 3: best_idx_interest * 3 + 3]
        best_fit_y = individual_gaussian(best_x, amp, cen, wid)
        best_x, best_fit_y = extrapolate_gaussian(best_x, amp, cen, wid, None, best_x.min() - 1, best_x.max() + 1, step=0.0001)
        new_ind_peak = (np.abs(best_x - x_full[ind_peak])).argmin()
        left_boundary, right_boundary = calculate_boundaries(best_x, best_fit_y, new_ind_peak, smoothing_params, pk_sns)
        best_x = best_x[left_boundary - 1: right_boundary + 1]
        best_fit_y = best_fit_y[left_boundary - 1: right_boundary + 1]
        area_smooth, area_ensemble = peak_area_distribution(best_fit_params, best_fit_params_error, best_idx_interest, best_x, x_full, ind_peak, multi=True, smoothing_params=smoothing_params, pk_sns=pk_sns)
    else:
        amp, cen, wid = best_fit_params[:3]
        tail_factor = 3
        x_min, x_max = calculate_gaus_extension_limits(cen, wid, 0, factor=tail_factor)
        # print("debug 1.6")
        if model_used == "asymmetric":
            alpha = best_fit_params[3]
            best_x, best_fit_y = extrapolate_gaussian(best_x, amp, cen, wid, alpha, x_min, x_max, step=0.0001)
            # print("debug 1.6.1")
        else:
            best_x, best_fit_y = extrapolate_gaussian(best_x, amp, cen, wid, None, x_min, x_max, step=0.0001)
            # print("debug 1.6.2")
        new_ind_peak = (np.abs(best_x - x_full[ind_peak])).argmin()
        # print("debug 1.7")
        left_boundary, right_boundary = calculate_boundaries_acceleration(best_x, best_fit_y, new_ind_peak, smoothing_params, pk_sns)
        # print("debug 1.8")
        area_smooth, area_ensemble = peak_area_distribution(best_fit_params, best_fit_params_error, best_idx_interest, best_x, x_full, ind_peak, multi=False, smoothing_params=smoothing_params, pk_sns=pk_sns)
        # print("debug 1.9")
    return best_x, best_fit_y, area_smooth, area_ensemble, best_result


def _fit_multi_gaussian(x_full, y_full, ind_peak, peaks, smoothing_params, pk_sns, gi):
    current_peaks = np.sort(np.append(peaks, ind_peak))
    best_fit_y = None
    best_fit_params = None
    best_fit_params_error = None
    best_x = None
    best_error = float("inf")
    best_idx_interest = None

    while len(current_peaks) > 1:
        left, _ = calculate_boundaries(x_full, y_full, np.min(current_peaks), smoothing_params, pk_sns)
        _, right = calculate_boundaries(x_full, y_full, np.max(current_peaks), smoothing_params, pk_sns)
        x = x_full[left:right + 1]
        y = y_full[left:right + 1]
        index_of_interest = np.where(current_peaks == ind_peak)[0][0]

        p0, bounds = [], ([], [])
        for peak in current_peaks:
            h, c, w = estimate_initial_gaussian_params(x, y, peak)
            p0.extend([h[0], c[0], w[0]])
            bounds[0].extend([0.1 * y_full[peak], x_full[peak] - 0.15, max(w[0] - 0.1, 0)])
            bounds[1].extend([1 + y_full[peak], x_full[peak] + 0.15, 0.5 + w[0]])

        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                popt, pcov = curve_fit(multigaussian, x, y, p0=p0, method="dogbox", bounds=bounds, maxfev=gi)
            fitted_y = multigaussian(x, *popt)
            error = np.sqrt(np.mean((fitted_y - y) ** 2))
            if error < best_error:
                best_error = error
                best_fit_params = popt
                best_fit_params_error = pcov
                best_fit_y = fitted_y
                best_x = x
                best_idx_interest = index_of_interest
        except RuntimeError:
            pass

        distances = np.abs(x[current_peaks] - x_full[ind_peak])
        if distances.size:
            current_peaks = np.delete(current_peaks, np.argmax(distances))

    if best_fit_params is not None:
        return best_x, best_fit_y, best_fit_params, best_fit_params_error, best_error, best_idx_interest
    return None

def _fit_single_gaussian(x_full, y_full, ind_peak, smoothing_params, pk_sns, gi, current_best_error):
    left, right = calculate_boundaries(x_full, y_full, ind_peak, smoothing_params, pk_sns)
    x = x_full[left:right + 1]
    y = y_full[left:right + 1]
    h, c, w = estimate_initial_gaussian_params(x, y, ind_peak)
    center_idx = (np.abs(x - c[0])).argmin()
    decay_init = estimate_initial_decay(x, y, center_idx)
    p0 = [h[0], c[0], w[0], decay_init]
    # p0 = [h[0], c[0], w[0], 0.1]
    bounds = ([0.9 * y_full[ind_peak], x_full[ind_peak] - 0.1, 0.5 * w[0], 0.01],
              [1 + y_full[ind_peak], x_full[ind_peak] + 0.1, 1.5 * w[0], 2])

    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            popt, pcov = curve_fit(gaussian_decay, x, y, p0=p0, method="dogbox", bounds=bounds, maxfev=gi)
        fitted_y = gaussian_decay(x, *popt)
        error = np.sqrt(np.mean((fitted_y - y) ** 2))
        if error < current_best_error:
            return x, fitted_y, popt, pcov, error
    except RuntimeError:
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                popt, pcov = curve_fit(gaussian_decay, x, y, p0=p0, method="dogbox", bounds=bounds, maxfev=gi * 1000)
            fitted_y = gaussian_decay(x, *popt)
            error = np.sqrt(np.mean((fitted_y - y) ** 2))
            if error < current_best_error:
                return x, fitted_y, popt, pcov, error
        except RuntimeError:
            tqdm.write("Error: Optimal parameters could not be found even after increasing iterations.")
    return None

def _fit_asymmetric_gaussian(x_full, y_full, ind_peak, smoothing_params, pk_sns, gi, current_best_error):
    left, right = calculate_boundaries(x_full, y_full, ind_peak, smoothing_params, pk_sns)
    x = x_full[left:right + 1]
    y = y_full[left:right + 1]

    h, c, w = estimate_initial_gaussian_params(x, y, ind_peak)

    # Sanitize estimates
    amp = max(h[0], 1e-5)
    cen = c[0]
    wid = max(w[0], 1e-5)
    alpha = 0.0  # Start symmetric

    p0 = [amp, cen, wid, alpha]
    bounds = (
        [1e-5, cen - 0.1, 1e-5, -10],  # lower bounds
        [10 * amp, cen + 0.1, 10 * wid, 10]  # upper bounds
    )

    try:
        popt, pcov = curve_fit(skewed_gaussian, x, y, p0=p0, bounds=bounds, maxfev=gi)
        fitted_y = skewed_gaussian(x, *popt)
        error = np.sqrt(np.mean((fitted_y - y) ** 2))
        if error < current_best_error:
            return x, fitted_y, popt, pcov, error
    except RuntimeError:
        pass
    return None

# def draw_positive_mvnorm(mu, cov, n_samples,max_decay = 15.0,  max_attempts = 100000):
#     """
#     Draws random fitting parameters from the covariance matrix while ensuring 
#     a positive width value, which is needed for peak integration. Maximum
#     decay value is 15 default. 

#     """
#     # out = []
#     # mu = np.asarray(mu)
#     # attempts = 0
#     # while len(out) < n_samples and attempts < max_attempts:
#     #     to_draw = n_samples - len(out)
#     #     batch = np.random.multivariate_normal(mu, cov, size=to_draw)
#     #     mask = batch[:,2] > 0
#     #     out.extend(batch[mask].tolist())
#     #     attempts += 1
#     # if len(out) < n_samples:
#     #     raise RuntimeError(f"Could only draw {len(out)} valid widths after {attempts} tries")
#     # return np.array(out[:n_samples])
def draw_positive_mvnorm(mu, cov, n_samples, max_attempts=10000):
    """
    Draw exactly n_samples from N(mu, cov) but only keep those with wid>0.
    We no longer filter on decay here.
    """
    mu = np.asarray(mu)
    out = []
    attempts = 0

    while len(out) < n_samples and attempts < max_attempts:
        to_draw = n_samples - len(out)
        batch = np.random.multivariate_normal(mu, cov, size=to_draw)
        # only require width > 0 (batch[:,2])
        mask = batch[:,2] > 0
        out.extend(batch[mask].tolist())
        attempts += 1

    if len(out) < n_samples:
        raise RuntimeError(
            f"Could only draw {len(out)} valid samples after {attempts} attempts "
            f"(needed {n_samples}).")

    return np.array(out[:n_samples])


def peak_area_distribution( params, params_uncertainty, ind, x, x_full, ind_peak, multi, smoothing_params, pk_sns, n_samples= 100):
    area_ensemble = []
    if multi:
        amp_i, cen_i, wid_i = params[ind * 3], params[ind * 3 + 1], params[ind * 3 + 2]
        start = 3*ind
        end = start+3
        pcov = params_uncertainty[start:end, start:end]
        # amp_unc_i, cen_unc_i, wid_unc_i = params_uncertainty[0], params_uncertainty[1], params_uncertainty[2]
        samples = np.random.multivariate_normal(np.array([amp_i, cen_i, wid_i]), pcov, size=n_samples)
        for i in range(0,n_samples):
            amp, cen, wid = samples[i,0], samples[i,1], samples[i,2]
            best_fit_y = individual_gaussian(x, amp, cen, wid)
            best_x, best_fit_y = extrapolate_gaussian(x, amp, cen, wid, None, x.min() - 1, x.max() + 1, step=0.0001)
            new_ind_peak = (np.abs(best_x - x_full[ind_peak])).argmin()
            left_boundary, right_boundary = calculate_boundaries(best_x, best_fit_y, new_ind_peak, smoothing_params, pk_sns)
            best_x = best_x[left_boundary - 1 : right_boundary + 1]
            best_fit_y = best_fit_y[left_boundary - 1 : right_boundary + 1]
            area_ensemble.append(simpson(y=best_fit_y, x=best_x))
        return np.median(area_ensemble), area_ensemble
    else:
        # samples = np.random.multivariate_normal(params, params_uncertainty, size=n_samples)
        # print("debug 1.8.1")
        # debug_param_distribution(params, params_uncertainty, n_draw=5000)
        # print("debug 1.8.1.5")
        samples = draw_positive_mvnorm(params, params_uncertainty,n_samples)
        # print("debug 1.8.2")
        x = 0
        for i in range(0,n_samples): 
            # print(f"1.8.2.{x}.a")
            amp, cen, wid, decay = samples[i]
            # print(f"1.8.2.{x}.b")
            decay_eff = np.clip(decay, 0.0, 15.0)
            wid   = max(abs(wid), 1e-6)
            decay = max(decay, 1e-6)
            # print(f"1.8.2.{x}.c")
            x_min, x_max = calculate_gaus_extension_limits(cen, wid, decay_eff, factor=3)
            # print(f"1.8.2.{x}.d")
            best_x, best_fit_y = extrapolate_gaussian(x, amp, cen, wid, decay_eff, x_min, x_max, step=1e-4)
            # x_min, x_max = calculate_gaus_extension_limits(cen, wid, decay, factor=3)
            # best_x, best_fit_y = extrapolate_gaussian(x, samples[i,0], samples[i,1], samples[i,2], samples[i,3], x_min, x_max, step=0.0001)
            new_ind_peak = (np.abs(best_x - x_full[ind_peak])).argmin()
            # print(f"1.8.2.{x}.e")
            left_boundary, right_boundary = calculate_boundaries(best_x, best_fit_y, new_ind_peak, smoothing_params, pk_sns)
            # print(f"1.8.2.{x}.f")
            best_x = best_x[left_boundary - 1 : right_boundary + 1]
            # print(f"1.8.2.{x}.g")
            best_fit_y = best_fit_y[left_boundary - 1 : right_boundary + 1]
            # print(f"1.8.2.{x}.h")
            area_ensemble.append(simpson(y=best_fit_y, x=best_x))
            # print(f"1.8.2.{x}.i")
            x=+1
        return np.median(area_ensemble), area_ensemble
    
def debug_param_distribution(mu, cov, n_draw=5000):
    """
    Sample from N(mu, cov) and plot:
      1) joint scatter of (wid, decay)
      2) histogram of wid
      3) histogram of decay
    """
    mu = np.asarray(mu)
    cov = np.asarray(cov)
    
    # draw a big batch
    batch = np.random.multivariate_normal(mu, cov, size=n_draw)
    wid   = batch[:,2]
    decay = batch[:,3]
    
    # print("Mean wid, decay:", mu[2], mu[3])
    # print("Std  wid, decay:", np.sqrt(cov[2,2]), np.sqrt(cov[3,3]))
    # print("   Sampled wid min/max:", wid.min(), wid.max())
    # print(" Sampled decay min/max:", decay.min(), decay.max())
    
    # 1) Joint scatter
    plt.figure()
    plt.scatter(wid, decay, alpha=0.2)
    plt.axvline(0)
    plt.axhline(0)
    plt.xlabel("wid")
    plt.ylabel("decay")
    plt.title("Joint draw of (wid, decay)")
    plt.show()
    
    # 2) wid histogram
    plt.figure()
    plt.hist(wid, bins=50)
    plt.xlabel("wid")
    plt.title("Histogram of wid")
    plt.show()
    
    # 3) decay histogram
    plt.figure()
    plt.hist(decay, bins=50)
    plt.xlabel("decay")
    plt.title("Histogram of decay")
    plt.show()
    
def individual_gaussian( x, amp, cen, wid):
    return amp * np.exp(-((x - cen) ** 2) / (2 * wid**2))

def estimate_initial_gaussian_params( x, y, peak):
    # Subset peaks so that only idx positions with x bounds are considered
    heights = []
    means = []
    stddevs = []
    height = y[peak]
    mean = x[peak]
    half_max = 0.5 * height
    mask = y >= half_max
    valid_x = x[mask]
    if len(valid_x) > 1:
        fwhm = np.abs(valid_x.iloc[-1] - valid_x.iloc[0])
        stddev = fwhm / (2 * np.sqrt(2 * np.log(2)))
    else:
        stddev = (x.max() - x.min()) / 6
    heights.append(height)
    means.append(mean)
    stddevs.append(stddev)
    return heights, means, stddevs

def estimate_initial_decay(x, y, center_idx):
    left_half = y[:center_idx]
    right_half = y[center_idx:]
    left_slope = np.mean(np.gradient(left_half))
    right_slope = np.mean(np.gradient(right_half))
    asymmetry = right_slope - left_slope

    # Empirical mapping to decay (tweak this based on real data behavior)
    decay_est = np.clip(0.5 * asymmetry, 0.01, 1.5)
    return decay_est

def multigaussian( x, *params):
    y = np.zeros_like(x)
    for i in range(0, len(params), 3):
        amp = params[i]
        cen = params[i + 1]
        wid = params[i + 2]
        y += amp * np.exp(-((x - cen) ** 2) / (2 * wid**2))
    return y

def skewed_gaussian(x, amp, cen, sigma, alpha):
    """
    Skewed Gaussian (Skew-Normal) distribution:
    - alpha = 0 gives symmetric Gaussian
    - alpha > 0 → right skew
    - alpha < 0 → left skew
    """
    z = (x - cen) / (sigma * np.sqrt(2))
    return amp * np.exp(-z**2) * (1 + erf(alpha * z))

def gaussian_decay( x, amp, cen, wid, dec):
    return amp * np.exp(-((x - cen) ** 2) / (2 * wid**2)) * np.exp(-dec * abs(x - cen))

def forward_derivative(x, y):
    fd = np.diff(y) / np.diff(x)
    x_n = x#[:-1]
    return fd, x_n

class FIDAnalyzer:
    def __init__(self, df, window_bounds, gaus_iterations, sample_name, is_reference, max_peaks, sw, sf, pk_sns, pk_pr, max_PA, reference_peaks=None):
        self.fig, self.axs = None, None
        self.df = df
        self.window_bounds = window_bounds
        self.sample_name = sample_name
        self.is_reference = is_reference
        self.reference_peaks = reference_peaks  # ref_key
        self.fig, self.axs = None, None
        self.datasets = []
        self.peaks_indices = []
        self.integrated_peaks = {}
        self.action_stack = []
        self.no_peak_lines = {}
        self.peaks = {}  # Store all peak indices and properties for each trace
        self.axs_to_traces = {}  # Empty map for connecting traces to figure axes
        self.peak_results = {}
        self.peak_results['Sample ID'] = sample_name
        self.gi = gaus_iterations
        self.max_peaks_for_neighborhood = max_peaks
        self.peak_properties = {}
        self.smoothing_params = [sw, sf]
        self.pk_sns = pk_sns
        self.pk_pr = pk_pr
        self.t_pressed = False # Flag to track if 't' was pressed
        self.called = False
        self.max_peak_amp = max_PA

    def run(self):
        """
        Executes the peak analysis workflow.
        Returns:
            peaks (dict): Peak areas and related info.
            fig (matplotlib.figure.Figure): The figure object.
            reference_peaks (dict): Updated reference peaks.
            t_pressed (bool): Indicates if 't' was pressed to update reference peaks.
        """
        self.fig, self.axs = self.plot_data()
        self.current_ax_idx = 0  # Initialize current axis index
        if self.is_reference:
            # Reference samples handling
            self.fig.canvas.mpl_connect("button_press_event", self.on_click)
            self.fig.canvas.mpl_connect("key_press_event", self.on_key)  # Connect general key events
            plt.show(block=True)  # Blocks script until plot window is closed
            if not self.reference_peaks:
                self.reference_peaks = self.peak_results
            else:
                self.reference_peaks.update(self.peak_results)
        else:
            # Non-reference samples handling
            self.auto_select_peaks()
            self.fig.canvas.mpl_connect("key_press_event", self.on_key)
            self.fig.canvas.mpl_connect("button_press_event", self.on_click)
            plt.show(block=True)  # Blocks script until plot window is closed
        return self.peak_results, self.fig, self.reference_peaks, self.t_pressed
    
def run_peak_integrator(data, key, gi, pk_sns, smoothing_params, max_peaks_for_neighborhood, fp, gaussian_fit_mode):
    # Setup data
    xdata = pd.Series(data['Samples'][key]['Raw Data'][data['Integration Metadata']['time_column']])
    ydata = pd.Series(data['Samples'][key]['Raw Data'][data['Integration Metadata']['signal_column']])
    
    # Subset to reference sample
    # --- Subset to global x-limits based on reference sample ---
    peak_times = list(data['Integration Metadata']['peak dictionary'].values())
    rt_buffer = 0.5  # 30 seconds = 0.5 minutes (you suggested 0.4, which is ~24s)
    
    xmin = min(peak_times) - rt_buffer
    xmax = max(peak_times) + rt_buffer
    mask = (xdata >= xmin) & (xdata <= xmax)
    
    xdata = xdata[mask].reset_index(drop=True)
    ydata = ydata[mask].reset_index(drop=True)
    
    # Signal processing
    ydata = smoother(ydata, smoothing_params[0], smoothing_params[1])
    ydata = pd.Series(ydata, index=xdata.index)
    ydata[ydata<0] = 0
    peak_timing = data['Integration Metadata']['peak dictionary'].values()
    data['Samples'][key]['Processed Data'] = {}
    
    base, min_peak_amp = baseline(xdata, ydata, deg=5, max_it=1000, tol=1e-4)
    y_bcorr = ydata-base
    peak_indices, peak_properties = find_peaks(y_bcorr, height=min_peak_amp, prominence=0.001)
    used_peaks = set()
    matched_indices = []
    presence_flags = []
    
    for pt in peak_timing:
        # Find candidate matches within tolerance
        distances = np.abs(xdata.iloc[peak_indices] - pt)
        candidates = [(idx, dist) for idx, dist in zip(peak_indices, distances) if dist <= 5/60]
    
        # Sort by closeness
        candidates.sort(key=lambda x: x[1])
    
        # Find the closest unused one
        selected = None
        for idx, dist in candidates:
            if idx not in used_peaks:
                selected = idx
                used_peaks.add(idx)
                break
    
        if selected is not None:
            matched_indices.append(selected)
            presence_flags.append(True)
        else:
            matched_indices.append(None)
            presence_flags.append(False)
    
    matched_indices = list(matched_indices)

    fig = plt.figure()
    plt.plot(xdata, y_bcorr, c= 'k', linewidth=1, linestyle='-', zorder=2)
    valleys = find_valleys(y_bcorr, peak_indices)
    peak_labels = list(data['Integration Metadata']['peak dictionary'])
    for label, peak_idx in zip(peak_labels, matched_indices):
        if peak_idx is None:          # in case some peaks weren’t matched
            data['Samples'][key]['Processed Data'][label] = [np.nan]
            continue
        try:
            if gaussian_fit_mode in {"multi", "both"}:
                A, B, peak_neighborhood = find_peak_neighborhood_boundaries(
                    x=xdata, y_smooth=y_bcorr, peaks=peak_indices, valleys=valleys,
                    peak_idx=peak_idx, max_peaks=max_peaks_for_neighborhood,
                    peak_properties=peak_properties, gi=gi,
                    smoothing_params=smoothing_params, pk_sns=pk_sns)
            else:
                peak_neighborhood = [peak_idx]
        
            x_fit, y_fit_smooth, area_smooth, area_ensemble, model_parameters = fit_gaussians(
                xdata, y_bcorr, peak_idx, peak_neighborhood,
                smoothing_params, pk_sns, gi=gi, mode=gaussian_fit_mode)
            plt.fill_between(x_fit, 0, y_fit_smooth, color="red", alpha=0.5, zorder=1)
            
            # Label
            x_peak_label = x_fit[np.argmax(y_fit_smooth)]
            y_peak_label = max(y_fit_smooth)
            plt.text(x_peak_label, y_peak_label * 1.05, label,
            ha='center', va='bottom',
            fontsize=8, color='black', rotation=0,
            zorder=2, bbox=dict(facecolor='white', edgecolor='none', alpha=0))
            # plt.axhline(0, c = 'k')
            
            # Assign data to output
            data['Samples'][key]['Processed Data'][label] = {
                 'Peak Area - median': np.median(area_ensemble),
                 'Peak Area - mean': np.mean(area_ensemble),
                 'Peak Area - standard deviation': np.std(area_ensemble, ddof=1),
                 'Peak Area - number of ensemble members': len(area_ensemble),
                 'Model Parameters': model_parameters,
                 'Retention Time': float(x_peak_label)}
        except Exception as e:
            tqdm.write(f"[Warning] Failed to fit {label} in {key}: {e}")
            data['Samples'][key]['Processed Data'][label] = [np.nan]
        
    
    peak_times = list(data['Integration Metadata']['peak dictionary'].values())
    mean_val = np.mean(peak_times)
    xmin = min(peak_times) - mean_val * 0.1
    xmax = max(peak_times) + mean_val * 0.1
    
    # new y max
    mask = (xdata >= xmin) & (xdata <= xmax)
    y_max = ydata[mask].max()
    plt.xlim(xmin, xmax)
    plt.ylim(0, y_max+y_max*0.1)
    plt.ylabel(data['Integration Metadata']['signal_column'])
    plt.xlabel(data['Integration Metadata']['time_column'])
    plt.savefig(str(fp)+f"/{key}.png", dpi=300)
    plt.close()
    return data





# ─── Standard Library ───────────────────────────────────────────────────────────
import os
import re
import sys
import shutil

# ─── Third-Party Libraries ─────────────────────────────────────────────────────
import numpy as np
import pandas as pd
import matplotlib
# matplotlib.use('Qt5Agg')
from matplotlib.widgets import TextBox, Button
import matplotlib.pyplot as plt
from scipy.signal import find_peaks
from tqdm import tqdm
import json

# ─── PyQt5 GUI Toolkit ─────────────────────────────────────────────────────────
from PyQt5.QtCore    import Qt
from PyQt5.QtWidgets import (
    QApplication,
    QDialog,
    QLineEdit,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QMessageBox)

# ─── Peak Integration ─────────────────────────────────────────────────────────
from .FID_Integration_functions import run_peak_integrator, smoother, baseline, find_valleys, find_peak_neighborhood_boundaries, fit_gaussians
    
def run_peak_integrator_manual(data, key, gi, pk_sns, smoothing_params, max_peaks_for_neighborhood, fp, gaussian_fit_mode):
    # Setup data
    md = data['Integration Metadata']
    x = pd.Series(data['Samples'][key]['Raw Data'][md['time_column']])
    x = x.fillna(0)
    y = pd.Series(data['Samples'][key]['Raw Data'][md['signal_column']])
    y = y.fillna(0)
    # Subset to x limits: either deduce from a dict of times, or fall back to explicit x-limits
    pdict = md['peak dictionary']
    if isinstance(pdict, dict):
         peak_times = list(pdict.values())
         labels     = list(pdict.keys())
         xmin, xmax = min(peak_times) - 0.4, max(peak_times) + 0.4
    else:
        # we're given just a list of labels
        labels     = pdict
        x0, x1     = md['x limits']
        xmin, xmax = x0 - 0.4, x1 + 0.4
    mask = (x >= xmin) & (x <= xmax)
    xdata = x[mask].reset_index(drop=True)
    ydata = y[mask].reset_index(drop=True)
    ydata = smoother(ydata, *smoothing_params)
    ydata = pd.Series(ydata, index=xdata.index)
    ydata[ydata < 0] = 0
    base, min_peak_amp = baseline(xdata, ydata, deg=500, max_it=1000, tol=1e-4)
    y_bcorr = ydata - base
    peak_indices, peak_properties = find_peaks(y_bcorr, height=min_peak_amp)#min_peak_amp)
    valleys = find_valleys(y_bcorr, peak_indices)
    peak_labels = labels

    # return data
    peak_selector = ManualPeakIntegrator(
        xdata,                   # as a numpy array
        y_bcorr,
        peak_indices,                       # pass the raw peak indices
        peak_properties,                    # needed for neighborhood fitting
        valleys,
        peak_labels,
        smoothing_params,
        pk_sns,
        gi,
        gaussian_fit_mode)
    # app = QApplication.instance() or QApplication(sys.argv)
    # app.exec_()
    app = QApplication.instance()
    owns_app = False
    if app is None:
        app = QApplication(sys.argv)
        owns_app = True
    
    if owns_app:
        app.exec_()
    else:
        while not getattr(peak_selector, "finished", False) and plt.fignum_exists(peak_selector.fig.number):
            plt.pause(0.1)
    
    
    if peak_selector.force_exit:
        tqdm.write("Manual integration was forcefully exited by the user.")
        raise SystemExit  # or return None, or raise a custom exception
    
    # Save output
    data['Samples'][key]['Processed Data'] = peak_selector.processed_data
    peak_selector.fig.savefig(str(fp) + f"/{key}.png", dpi=300)
    plt.close(peak_selector.fig)
    return data


class ManualPeakIntegrator:
    def __init__(self,
                 x, y,
                 peaks,           # <— list/array of peak indices
                 peak_properties,
                 valleys,
                 labels,
                 smoothing_params,
                 pk_sns,
                 gi,
                 gaussian_fit_mode,
                 owns_app=False):
        self._owns_app = owns_app
        self.x, self.y = pd.Series(x), pd.Series(y)
        self.valleys = valleys
        self.peaks = np.asarray(peaks)
        self.peak_properties = peak_properties
        self.labels = labels
        self.smoothing_params = smoothing_params
        self.pk_sns = pk_sns
        self.gi = gi
        self.gaussian_fit_mode = gaussian_fit_mode

        self.index = 0
        self.processed_data = {}
        self.artists_stack = []
        self.click_tolerance = 10/60  # 10 seconds in minutes
        self.finished = False

        # figure + data plot
        self.fig, self.ax = plt.subplots()
        self.ax.axhline(0, c='k')
        self.ax.plot(self.x, self.y, c='k', alpha = 0.6)
        self.text = self.ax.text(
            0.5, 0.95, f"Click peak for: {self.labels[self.index]}",
            transform=self.ax.transAxes, ha='center')
        self.ax.set_xlabel('Retention Time (min)')
        self.ax.set_ylabel("Value (pA)")
        # “Finished” button
        btn_ax = self.fig.add_axes([0.82, 0.02, 0.15, 0.05])
        self.finish_button = Button(btn_ax, "Finished")
        # Now finish accepts the event
        self.finish_button.on_clicked(self.finish)
        
        # Exit button
        exit_ax = self.fig.add_axes([0.82, 0.94, 0.15, 0.05])
        self.exit_button = Button(exit_ax, "Exit")
        self.exit_button.on_clicked(self.exit_program)
        
        # Flag for termination
        self.force_exit = False

        # connect events
        self.cid_click = self.fig.canvas.mpl_connect("button_press_event", self.onclick)
        self.cid_key   = self.fig.canvas.mpl_connect("key_press_event",   self.on_key)

    def exit_program(self, event=None):
        """Triggered when the Exit button is clicked."""
        self.force_exit = True
        self.text.set_text("Exiting...")
        self.fig.canvas.draw()
        # QApplication.quit()
        if getattr(self, "_owns_app", False):
            app = QApplication.instance()
            if app is not None:
                app.quit()
        plt.close(self.fig)
        
    # def onclick(self, event):
    #     if event.inaxes != self.ax:
    #         return
    #     # Check if peaks are selected
    #     if self.index >= len(self.labels):
    #        msg = "[Manual] All peaks have already been selected. No more selections expected."
    #        try:
    #            tqdm.write(msg)
    #        except Exception:
    #            print(msg)
    #        # (optional) stop processing further clicks
    #        try:
    #            self.fig.canvas.mpl_disconnect(self.cid_click)
    #        except Exception:
    #            pass
    #        return
        
    #     click_time = event.xdata
    #     peak_times = self.x.to_numpy()[self.peaks]
    #     dists = np.abs(peak_times - click_time)
    #     best = dists.argmin()
   
    #     # if the nearest real peak is > tolerance, treat as “no peak”
    #     if dists[best] <= self.click_tolerance:
    #         peak_idx = int(self.peaks[best])
    #     else:
    #         # no valid peak → grey dashed line & record NaN
    #         line = self.ax.axvline(click_time, color='grey', linestyle='--')
    #         self.artists_stack.append([line])
    #         self.processed_data[self.labels[self.index]] = {'Values': [np.nan]}
    #         self._advance_prompt()
    #         return
    #     drawn = []
    #     try:
    #         if self.gaussian_fit_mode in {"multi","both"}:
    #             _, _, neigh = find_peak_neighborhood_boundaries(
    #                 self.x, self.y, self.peaks, self.valleys,
    #                 peak_idx, self.pk_sns,
    #                 peak_properties=self.peak_properties,
    #                 gi=self.gi,
    #                 smoothing_params=self.smoothing_params,
    #                 pk_sns=self.pk_sns)
    #         else:
    #             neigh = [peak_idx]
    #         # print("debug 1")   
    #         x_fit, y_fit, _, area_ensemble, model_params = fit_gaussians(
    #             self.x, self.y, peak_idx, neigh,
    #             self.smoothing_params, self.pk_sns,
    #             gi=self.gi,
    #             mode=self.gaussian_fit_mode)
    #         # print("debug 2")
    #         poly = self.ax.fill_between(x_fit, 0, y_fit, color='red', alpha=0.4)
    #         drawn.append(poly)
    #         self.processed_data[self.labels[self.index]] = {
    #             'Peak Area - median': np.median(area_ensemble),
    #             'Peak Area - mean': np.mean(area_ensemble),
    #             'Peak Area - standard deviation': np.std(area_ensemble, ddof=1),
    #             'Peak Area - number of ensemble members': len(area_ensemble),
    #             'Model Parameters': model_params,
    #             'Retention Time': float(click_time)}
   
    #     except Exception as e:
    #         tqdm.write(f"[Manual Warning] Failed to fit {self.labels[self.index]}: {e}")
    #         line = self.ax.axvline(click_time, color='grey', linestyle='--')
    #         drawn.append(line)
    #         self.processed_data[self.labels[self.index]] = {'Values':[np.nan]}
   
    #     # save for undo, then advance
    #     self.artists_stack.append(drawn)
    #     self._advance_prompt()
    def onclick(self, event):
        if event.inaxes != self.ax:
            return
    
        # ---- Guard: all expected peaks already selected ----
        if self.index >= len(self.labels):
            msg = "[Manual] All peaks have already been selected. No more selections expected."
            try:
                tqdm.write(msg)
            except Exception:
                print(msg)
            return
    
        # Safe current label (use after the guard above)
        current_label = self.labels[self.index]
    
        click_time = event.xdata
        peak_times = self.x.to_numpy()[self.peaks]
        dists = np.abs(peak_times - click_time)
        best = dists.argmin()
    
        # if the nearest real peak is > tolerance, treat as “no peak”
        if dists[best] > self.click_tolerance:
            # no valid peak → grey dashed line & record NaN
            line = self.ax.axvline(click_time, color='grey', linestyle='--')
            self.artists_stack.append([line])
            self.processed_data[current_label] = {'Values': [np.nan]}
            self._advance_prompt()
            return
    
        peak_idx = int(self.peaks[best])
    
        drawn = []
        try:
            if self.gaussian_fit_mode in {"multi", "both"}:
                _, _, neigh = find_peak_neighborhood_boundaries(
                    self.x, self.y, self.peaks, self.valleys,
                    peak_idx, self.pk_sns,
                    peak_properties=self.peak_properties,
                    gi=self.gi,
                    smoothing_params=self.smoothing_params,
                    pk_sns=self.pk_sns
                )
            else:
                neigh = [peak_idx]
    
            x_fit, y_fit, _, area_ensemble, model_params = fit_gaussians(
                self.x, self.y, peak_idx, neigh,
                self.smoothing_params, self.pk_sns,
                gi=self.gi,
                mode=self.gaussian_fit_mode)
    
            poly = self.ax.fill_between(x_fit, 0, y_fit, color='red', alpha=0.4)
            drawn.append(poly)
    
            self.processed_data[current_label] = {
                'Peak Area - median': float(np.median(area_ensemble)),
                'Peak Area - mean': float(np.mean(area_ensemble)),
                'Peak Area - standard deviation': float(np.std(area_ensemble, ddof=1)),
                'Peak Area - number of ensemble members': int(len(area_ensemble)),
                'Model Parameters': model_params,
                'Retention Time': float(click_time),
            }
    
        except Exception as e:
            # Don't index self.labels[self.index] here – use current_label captured earlier
            # try:
            #     tqdm.write(f"[Manual Warning] Failed to fit {current_label}: {e}")
            # except Exception:
            #     print(f"[Manual Warning] Failed to fit {current_label}: {e}")
            tqdm.write(f"[Manual Warning] Failed to fit {current_label}: {e}")
            line = self.ax.axvline(click_time, color='grey', linestyle='--')
            drawn.append(line)
            self.processed_data[current_label] = {'Values': [np.nan]}
    
        # save for undo, then advance
        self.artists_stack.append(drawn)
        self._advance_prompt()
    
        # If that was the last label, optionally inform user & disconnect clicks
        if self.index >= len(self.labels):
            done_msg = "[Manual] All peaks selected. You can press Finished."
            try:
                tqdm.write(done_msg)
            except Exception:
                print(done_msg)
            try:
                self.fig.canvas.mpl_disconnect(self.cid_click)
            except Exception:
                pass
    
    def _advance_prompt(self):
        """increment index, update the onscreen prompt, redraw."""
        self.index += 1
        if self.index < len(self.labels):
            self.text.set_text(f"Click peak for: {self.labels[self.index]}")
        else:
            self.text.set_text("All peaks selected. Click 'Finished' to proceed.")
        self.fig.canvas.draw_idle()

    def on_key(self, event):
        key = event.key.lower()
        if key in ('shift+delete','shift+del','shift+backspace') and self.index > 0:
            # undo
            self.index -= 1
            label = self.labels[self.index]
            self.processed_data.pop(label, None)
            last = self.artists_stack.pop()
            for art in last:
                try: art.remove()
                except: pass
            self.text.set_text(f"Click peak for: {self.labels[self.index]}")
            self.fig.canvas.draw()

    def finish(self, event=None):
        # disconnect callbacks and close GUI
        self.text.set_text("")
        self.fig.canvas.draw
        self.fig.canvas.mpl_disconnect(self.cid_click)
        self.fig.canvas.mpl_disconnect(self.cid_key)
        self.finished=True
        # QApplication.quit()
        if getattr(self, "_owns_app", False):
            app = QApplication.instance()
            if app is not None:
                app.quit()


import os
import shutil
from PIL import Image
import imagehash
import numpy as np
import matplotlib.pyplot as plt
from sklearn.manifold import MDS
from sklearn.mixture import GaussianMixture
from scipy.cluster.hierarchy import linkage, dendrogram
from .FID_integration import import_data, save_json
from .FID_General import plot_chromatogram, plot_chromatogram_cluster

def cluster(time_window=None, max_clusters = 4, copy_files = True):
    data, chrom_folder , output_path, folder_path = plot_chromatogram(time_window)
    data = clusterer(data, chrom_folder, max_clusters=max_clusters)
    if copy_files:
        cluster_files(data, folder_path)
    save_json({"data_dict": data}, output_path)

def clusterer(data, chromatogram_folder, max_clusters):#, dendro_diagram=False):
    """
    Cluster chromatograms using perceptual hashing and determine optimal number
    of clusters via BIC on Gaussian Mixture Models.

    Parameters
    ----------
    data : dict
        Dictinoary data structure produced using import_data().
    chromatogram_folder : str
        Pathway to folder containing chromatograms for clustering via perceptual hashing.
    max_clusters : int
        Maximum number of clusters to split data into. Default 4.
    dendro_diagram: bool
        Figure displaying the number of clusters and assocaited BIC. Defualt False.

    Returns
    -------
    None.

    """
    # Load chromatogram images
    image_files = [f for f in os.listdir(chromatogram_folder) if f.lower().endswith(('.png', '.jpg', '.jpeg', '.bmp'))]
    image_paths = [os.path.join(chromatogram_folder, f) for f in image_files]
    sample_names = [os.path.splitext(f)[0] for f in image_files]
    
    def get_perceptual_hash(image_path):
        img = Image.open(image_path).convert("L")
        return imagehash.phash(img)
    
    hashes = [get_perceptual_hash(path) for path in image_paths]

    # Compute Hamming distance matrix
    n = len(hashes)
    dist_matrix = np.zeros((n, n))
    for i in range(n):
        for j in range(i + 1, n):
            dist = hashes[i] - hashes[j]
            dist_matrix[i, j] = dist_matrix[j, i] = dist

    # Dimensionality reduction using MDS
    mds = MDS(n_components=4, dissimilarity='precomputed', random_state=42)
    embedding = mds.fit_transform(dist_matrix)

    # BIC-based selection of optimal number of clusters
    lowest_bic = np.infty
    best_gmm = None
    bic_scores = []

    for k in range(2, max_clusters + 1):
        gmm = GaussianMixture(n_components=k, covariance_type='full', random_state=42)
        gmm.fit(embedding)
        bic = gmm.bic(embedding)
        bic_scores.append(bic)
        if bic < lowest_bic:
            lowest_bic = bic
            best_gmm = gmm
    print(f"Optimal number of clusters identified by BIC: {best_gmm.n_components}")

    optimal_clusters = best_gmm.n_components
    cluster_labels = best_gmm.predict(embedding)

    # BIC plot
    plt.figure()
    plt.plot(range(2, max_clusters + 1), bic_scores, marker='o')
    plt.xlabel("Number of Clusters")
    plt.ylabel("BIC")
    plt.show()

    for name, group in zip(sample_names, cluster_labels):
        data['Samples'][name]['cluster'] = int(group + 1)  # 1-based indexing
    return data

def cluster_files(data, folder_path):
    assignments = {
        name: sample_info["cluster"]
        for name, sample_info in data["Samples"].items()}
    
    for sample_name, cluster_num in assignments.items():
        src = os.path.join(folder_path, f"{sample_name}.txt")
        dest_dir = os.path.join(folder_path, f"Cluster {cluster_num}")
        os.makedirs(dest_dir, exist_ok=True)
        try:
            shutil.copy(src, dest_dir)
        except FileNotFoundError:
            print(f"⚠️  raw file not found: {src}")
        # else:
        #     print(f"→ Copied {sample_name}.txt to Cluster {cluster_num}/")